import pytest

from ase.build import molecule

from gpaw.mpi import world
from gpaw import GPAW
from gpaw.lrtddft.kssingle import KSSingles


@pytest.mark.lrtddft
def test_old_io(in_tmp_dir):
    """Test reading of old style output files"""
    fname = 'veryold.dat'
    if world.rank == 0:
        with open(fname, 'w') as f:
            f.write("""# KSSingles
2
0 1   0 0   0.129018 1   -4.7624e-02 -3.2340e-01 -4.6638e-01
0 1   1 0   0.129018 1   -4.7624e-02 -3.2340e-01 -4.6638e-01
""")
    world.barrier()

    kss = KSSingles.read(fname)
    assert len(kss) == 2

    fname = 'old.dat'
    if world.rank == 0:
        with open(fname, 'w') as f:
            f.write("""# KSSingles
2 float64
0.024
0 1  0 0  0.392407 2  8.82 2.91 7.98  1.0 2.7 7.8   1.4 1.2 1.08
0 2  0 0  0.402312 2  2.24 8.49 -5.24   2.3 8.26 -4.99038e-14
""")
    world.barrier()

    kss = KSSingles.read(fname)
    assert len(kss) == 2
    assert kss.restrict['eps'] == 0.024


@pytest.fixture
def ch4():
    ch4 = molecule('CH4')
    ch4.center(vacuum=2)
    ch4.calc = GPAW(h=0.25, nbands=8, txt=None)
    ch4.get_potential_energy()
    return ch4


@pytest.mark.lrtddft
def test_io(in_tmp_dir, ch4):
    # full KSSingles
    kssfull = KSSingles(restrict={'eps': 0.9})
    kssfull.calculate(ch4)
    fullname = 'kssfull.dat'
    kssfull.write(fullname)
    world.barrier()

    # read full
    kss1 = KSSingles.read(fullname)
    assert len(kss1) == 16

    # restricted KSSingles
    istart, jend = 1, 4
    kss = KSSingles(restrict={'eps': 0.9, 'istart': istart, 'jend': jend})
    kss.calculate(ch4)
    f14name = 'kss_1_4.dat'
    kss.write(f14name)
    world.barrier()

    kss2 = KSSingles.read(f14name)
    assert len(kss2) == len(kss)
    assert kss2.restrict['istart'] == istart
    assert kss2.restrict['jend'] == jend

    # restrict when reading
    kss3 = KSSingles.read(fullname,
                          restrict={'istart': istart, 'jend': jend})
    assert len(kss3) == len(kss)
    assert kss3.restrict['istart'] == istart
    assert kss3.restrict['jend'] == jend


@pytest.mark.lrtddft
def test_mul(ch4):
    """Test multiplication with a number"""
    istart, jend = 1, 4
    kss = KSSingles(restrict={'eps': 0.9, 'istart': istart, 'jend': jend})
    kss.calculate(ch4)

    ks0 = kss[0]
    ks1 = 1. * ks0
    assert (ks0.me == ks1.me).all()
    assert (ks0.magn == ks1.magn).all()



from ase import Atoms
from gpaw import GPAW, PW
from gpaw.mpi import world, serial_comm


def test_pw_fulldiag(in_tmp_dir, scalapack):
    a = Atoms('H2',
              [(0, 0, 0), (0, 0, 0.74)],
              cell=(3, 3, 3),
              pbc=1)

    a.calc = GPAW(mode=PW(force_complex_dtype=True),
                  eigensolver='rmm-diis',
                  nbands=8,
                  parallel={'domain': 1},
                  basis='dzp',
                  txt='H2.txt')

    a.get_potential_energy()
    _ = a.calc.get_pseudo_wave_function(0)
    e1 = a.calc.get_eigenvalues()
    w1 = a.calc.get_pseudo_wave_function(0)

    a.calc.write('H2.gpw')

    if world.size == 1:
        scalapack = None
    else:
        scalapack = (2, world.size // 2, 32)

    a.calc.diagonalize_full_hamiltonian(nbands=120, scalapack=scalapack)
    w2 = a.calc.get_pseudo_wave_function(0)
    e2 = a.calc.get_eigenvalues()

    calc = GPAW('H2.gpw', txt=None, parallel={'domain': 1})
    calc.diagonalize_full_hamiltonian(nbands=120, scalapack=scalapack)
    w3 = calc.get_pseudo_wave_function(0)
    e3 = calc.get_eigenvalues()

    calc.write('H2wf.gpw', 'all')

    calc = GPAW('H2wf.gpw', txt=None, communicator=serial_comm)
    w4 = calc.get_pseudo_wave_function(0)
    e4 = calc.get_eigenvalues()

    for w in [w2, w3, w4]:
        err = abs(abs(w[1, 2, 3]) - abs(w1[1, 2, 3]))
        assert err < 1e-7, err

    for e in [e2, e3, e4]:
        err = abs(e[1] - e1[1])
        assert err < 1e-9, err
        err = abs(e[-1] - e2[-1])
        assert err < 1e-10, err



from gpaw.solvation.gridmem import NeedsGD
from ase.units import Bohr, Hartree
import numpy as np


class Interaction(NeedsGD):
    """Base class for non-electrostatic solvent solute interactions.

    See also Section III of
    A. Held and M. Walter, J. Chem. Phys. 141, 174108 (2014).

    Attributes:
    subscript         -- Short label used to reference the interaction.
    E                 -- Local interaction energy in Hartree
    delta_E_delta_n_g -- Functional derivative of the interaction energy
                         with respect to the electron density.
    delta_E_delta_n_g -- Functional derivative of the interaction energy
                         with respect to the cavity.
    """

    subscript = 'unnamed'

    def __init__(self):
        NeedsGD.__init__(self)
        self.E = None
        self.delta_E_delta_n_g = None
        self.delta_E_delta_g_g = None

    def write(self, writer):
        pass

    def update(self, atoms, density, cavity):
        """Update the Kohn-Sham potential and the energy.

        atoms and/or cavity are None iff they have not changed
        since the last call

        Return whether the interaction has changed.
        """
        raise NotImplementedError

    def estimate_memory(self, mem):
        ngrids = 1 + self.depends_on_el_density
        mem.subnode('Functional Derivatives', ngrids * self.gd.bytecount())

    def allocate(self):
        NeedsGD.allocate(self)
        self.delta_E_delta_g_g = self.gd.empty()
        if self.depends_on_el_density:
            self.delta_E_delta_n_g = self.gd.empty()

    @property
    def depends_on_atomic_positions(self):
        """Return whether the ia depends explicitly on atomic positions."""
        raise NotImplementedError

    @property
    def depends_on_el_density(self):
        """Return whether the ia depends explicitly on the electron density."""
        raise NotImplementedError

    def get_del_r_vg(self, atomic_index, density):
        """Return spatial derivatives with respect to atomic position."""
        raise NotImplementedError

    def __str__(self):
        s = "Interaction: %s\n" % (self.__class__, )
        s += "  subscript: %s\n" % (self.subscript, )
        return s

    def update_atoms(self, atoms, log):
        """Inexpensive initialization when atoms change."""
        pass


class SurfaceInteraction(Interaction):
    """An interaction with energy proportional to the cavity surface area."""

    subscript = 'surf'
    depends_on_el_density = False
    depends_on_atomic_positions = False

    def __init__(self, surface_tension):
        """Constructor for SurfaceInteraction class.

        Arguments:
        surface_tension -- Proportionality factor to calculate
                           energy from surface area in eV / Angstrom ** 2.
        """
        Interaction.__init__(self)
        self.surface_tension = float(surface_tension)

    def write(self, writer):
        writer.write(name='SurfaceInteraction',
                     surface_tension=self.surface_tension)

    def update(self, atoms, density, cavity):
        if cavity is None:
            return False
        acalc = cavity.surface_calculator
        st = self.surface_tension * Bohr ** 2 / Hartree
        self.E = st * acalc.A
        np.multiply(st, acalc.delta_A_delta_g_g, self.delta_E_delta_g_g)
        return True

    def __str__(self):
        s = Interaction.__str__(self)
        s += '  surface_tension: %s\n' % (self.surface_tension, )
        return s


class VolumeInteraction(Interaction):
    """An interaction with energy proportional to the cavity volume"""

    subscript = 'vol'
    depends_on_el_density = False
    depends_on_atomic_positions = False

    def __init__(self, pressure):
        """Constructor for VolumeInteraction class.

        Arguments:
        pressure -- Proportionality factor to calculate
                    energy from volume in eV / Angstrom ** 3.
        """
        Interaction.__init__(self)
        self.pressure = float(pressure)

    def update(self, atoms, density, cavity):
        if cavity is None:
            return False
        vcalc = cavity.volume_calculator
        pressure = self.pressure * Bohr ** 3 / Hartree
        self.E = pressure * vcalc.V
        np.multiply(pressure, vcalc.delta_V_delta_g_g, self.delta_E_delta_g_g)
        return True

    def __str__(self):
        s = Interaction.__str__(self)
        s += '  pressure: %s\n' % (self.pressure, )
        return s


class LeakedDensityInteraction(Interaction):
    """Interaction proportional to charge leaking outside cavity.

    The charge outside the cavity is calculated as
    """

    subscript = 'leak'
    depends_on_el_density = True
    depends_on_atomic_positions = False

    def __init__(self, voltage):
        """Constructor for LeakedDensityInteraction class.

        Arguments:
        voltage -- Proportionality factor to calculate
                   energy from integrated electron density in Volts.
                   A positive value of the voltage leads to a
                   positive interaction energy.
        """
        Interaction.__init__(self)
        self.voltage = float(voltage)

    def update(self, atoms, density, cavity):
        E0 = self.voltage / Hartree
        if cavity is not None:
            np.multiply(E0, cavity.g_g, self.delta_E_delta_n_g)
        np.multiply(E0, density.nt_g, self.delta_E_delta_g_g)
        self.E = self.gd.integrate(
            density.nt_g * self.delta_E_delta_n_g,
            global_integral=False
        )
        return True

    def __str__(self):
        s = Interaction.__str__(self)
        s += '  voltage: %s\n' % (self.voltage, )
        return s



from ase.build import molecule

from gpaw.test import equal
import gpaw.solvation as solv
from gpaw.cluster import Cluster
from gpaw.lrtddft import LrTDDFT
from gpaw import PoissonSolver


def test_solvation_lrtddft():
    h = 0.3
    vac = 3.0

    atoms = Cluster(molecule('H2'))
    atoms.minimal_box(vac, h)

    calc = solv.SolvationGPAW(
        xc='PBE', h=0.2,  # non-solvent DFT parameters
        nbands=3,
        convergence={'energy': 0.1, 'eigenstates': 0.01, 'density': 0.1},
        # convenient way to use HW14 water parameters:
        **solv.get_HW14_water_kwargs())

    # do the ground state calculation
    atoms.calc = calc
    atoms.get_potential_energy()
    print(id(calc.hamiltonian.poisson.dielectric))
    print(id(calc.hamiltonian.dielectric))

    # linear response using ground state Poisson
    lrw = LrTDDFT(calc)
    lrw.diagonalize()

    # We test the agreement of a pure RPA kernel
    # with setting eps to 1
    
    lr = LrTDDFT(calc,
                 poisson=PoissonSolver('fd', nn=calc.hamiltonian.poisson.nn))
    lr.diagonalize()

    calc.hamiltonian.poisson.dielectric.epsinf = 1.
    lr1 = LrTDDFT(calc)
    lr1.diagonalize()
    for ex, ex1 in zip(lr, lr1):
        equal(ex.energy, ex1.energy, 1e-14)



import numpy as np
from ase.units import Ha

from gpaw.projections import Projections
from gpaw.utilities import pack, unpack2
from gpaw.utilities.blas import axpy, mmm
from gpaw.utilities.partition import AtomPartition


class WaveFunctions:
    """...

    setups:
        List of setup objects.
    symmetry:
        Symmetry object.
    kpt_u:
        List of **k**-point objects.
    nbands: int
        Number of bands.
    nspins: int
        Number of spins.
    dtype: dtype
        Data type of wave functions (float or complex).
    bzk_kc: ndarray
        Scaled **k**-points used for sampling the whole
        Brillouin zone - values scaled to [-0.5, 0.5).
    ibzk_kc: ndarray
        Scaled **k**-points in the irreducible part of the
        Brillouin zone.
    weight_k: ndarray
        Weights of the **k**-points in the irreducible part
        of the Brillouin zone (summing up to 1).
    kpt_comm:
        MPI-communicator for parallelization over **k**-points.
    """

    def __init__(self, gd, nvalence, setups, bd, dtype, collinear,
                 world, kd, kptband_comm, timer):
        self.gd = gd
        self.nspins = kd.nspins
        self.collinear = collinear
        self.nvalence = nvalence
        self.bd = bd
        self.dtype = dtype
        assert dtype == float or dtype == complex
        self.world = world
        self.kd = kd
        self.kptband_comm = kptband_comm
        self.timer = timer
        self.atom_partition = None

        self.kpt_qs = kd.create_k_points(self.gd.sdisp_cd, collinear)
        self.kpt_u = [kpt for kpt_s in self.kpt_qs for kpt in kpt_s]

        self.occupations = None
        self.fermi_levels = None

        self.eigensolver = None
        self.positions_set = False
        self.spos_ac = None

        self.set_setups(setups)

    @property
    def fermi_level(self):
        assert len(self.fermi_levels) == 1
        return self.fermi_levels[0]

    def summary(self, log):
        log(eigenvalue_string(self))

        if self.fermi_levels is None:
            return

        if len(self.fermi_levels) == 1:
            log(f'Fermi level: {self.fermi_levels[0] * Ha:.5f}\n')
        else:
            f1, f2 = (f * Ha for f in self.fermi_levels)
            log(f'Fermi levels: {f1:.5f}, {f2:.5f}\n')

    def set_setups(self, setups):
        self.setups = setups

    def set_eigensolver(self, eigensolver):
        self.eigensolver = eigensolver

    def add_realspace_orbital_to_density(self, nt_G, psit_G):
        if psit_G.dtype == float:
            axpy(1.0, psit_G**2, nt_G)
        else:
            assert psit_G.dtype == complex
            axpy(1.0, psit_G.real**2, nt_G)
            axpy(1.0, psit_G.imag**2, nt_G)

    def add_orbital_density(self, nt_G, kpt, n):
        self.add_realspace_orbital_to_density(nt_G, kpt.psit_nG[n])

    def calculate_band_energy(self):
        e_band = 0.0
        for kpt in self.kpt_u:
            e_band += np.dot(kpt.f_n, kpt.eps_n)

        try:  # DCSF needs this ...
            e_band += self.occupations.calculate_band_energy(self)
        except AttributeError:
            pass

        return self.kptband_comm.sum(e_band)

    def calculate_density_contribution(self, nt_sG):
        """Calculate contribution to pseudo density from wave functions.

        Array entries are written to (not added to)."""
        nt_sG.fill(0.0)
        for kpt in self.kpt_u:
            self.add_to_density_from_k_point(nt_sG, kpt)
        self.kptband_comm.sum(nt_sG)

        self.timer.start('Symmetrize density')
        for nt_G in nt_sG:
            self.kd.symmetry.symmetrize(nt_G, self.gd)
        self.timer.stop('Symmetrize density')

    def add_to_density_from_k_point(self, nt_sG, kpt):
        self.add_to_density_from_k_point_with_occupation(nt_sG, kpt, kpt.f_n)

    def get_orbital_density_matrix(self, a, kpt, n):
        """Add the nth band density from kpt to density matrix D_sp"""
        ni = self.setups[a].ni
        D_sii = np.zeros((self.nspins, ni, ni))
        P_i = kpt.P_ani[a][n]
        D_sii[kpt.s] += np.outer(P_i.conj(), P_i).real
        D_sp = [pack(D_ii) for D_ii in D_sii]
        return D_sp

    def calculate_atomic_density_matrices_k_point(self, D_sii, kpt, a, f_n):
        if kpt.rho_MM is not None:
            P_Mi = self.P_aqMi[a][kpt.q]
            rhoP_Mi = np.zeros_like(P_Mi)
            D_ii = np.zeros(D_sii[kpt.s].shape, kpt.rho_MM.dtype)
            mmm(1.0, kpt.rho_MM, 'N', P_Mi, 'N', 0.0, rhoP_Mi)
            mmm(1.0, P_Mi, 'C', rhoP_Mi, 'N', 0.0, D_ii)
            D_sii[kpt.s] += D_ii.real
        else:
            if self.collinear:
                P_ni = kpt.projections[a]
                D_sii[kpt.s] += np.dot(P_ni.T.conj() * f_n, P_ni).real
            else:
                P_nsi = kpt.projections[a]
                D_ssii = np.einsum('nsi,n,nzj->szij',
                                   P_nsi.conj(), f_n, P_nsi)
                D_sii[0] += (D_ssii[0, 0] + D_ssii[1, 1]).real
                D_sii[1] += 2 * D_ssii[0, 1].real
                D_sii[2] += 2 * D_ssii[0, 1].imag
                D_sii[3] += (D_ssii[0, 0] - D_ssii[1, 1]).real

        if hasattr(kpt, 'c_on'):
            for ne, c_n in zip(kpt.ne_o, kpt.c_on):
                d_nn = ne * np.outer(c_n.conj(), c_n)
                D_sii[kpt.s] += np.dot(P_ni.T.conj(), np.dot(d_nn, P_ni)).real

    def calculate_atomic_density_matrices(self, D_asp):
        """Calculate atomic density matrices from projections."""
        f_un = [kpt.f_n for kpt in self.kpt_u]
        self.calculate_atomic_density_matrices_with_occupation(D_asp, f_un)

    def calculate_atomic_density_matrices_with_occupation(self, D_asp, f_un):
        """Calculate atomic density matrices from projections with
        custom occupation f_un."""

        # Parameter check (if user accidentally passes f_n instead of f_un)
        if f_un[0] is not None:  # special case for transport calculations...
            assert isinstance(f_un[0], np.ndarray)
        # Varying f_n used in calculation of response part of GLLB-potential
        for a, D_sp in D_asp.items():
            ni = self.setups[a].ni
            D_sii = np.zeros((len(D_sp), ni, ni))
            for f_n, kpt in zip(f_un, self.kpt_u):
                self.calculate_atomic_density_matrices_k_point(D_sii, kpt, a,
                                                               f_n)
            D_sp[:] = [pack(D_ii) for D_ii in D_sii]
            self.kptband_comm.sum(D_sp)

        self.symmetrize_atomic_density_matrices(D_asp)

    def symmetrize_atomic_density_matrices(self, D_asp):
        if len(self.kd.symmetry.op_scc) == 0:
            return

        a_sa = self.kd.symmetry.a_sa
        D_asp.redistribute(self.atom_partition.as_serial())
        for s in range(self.nspins):
            D_aii = [unpack2(D_asp[a][s])
                     for a in range(len(D_asp))]
            for a, D_ii in enumerate(D_aii):
                setup = self.setups[a]
                D_asp[a][s] = pack(setup.symmetrize(a, D_aii, a_sa))
        D_asp.redistribute(self.atom_partition)

    def calculate_occupation_numbers(self, fixed_fermi_level=False):
        if self.collinear and self.nspins == 1:
            degeneracy = 2
        else:
            degeneracy = 1

        f_qn, fermi_levels, e_entropy = self.occupations.calculate(
            nelectrons=self.nvalence / degeneracy,
            eigenvalues=[kpt.eps_n * Ha for kpt in self.kpt_u],
            weights=[kpt.weightk for kpt in self.kpt_u],
            fermi_levels_guess=self.fermi_levels * Ha
            if self.fermi_levels is not None else None)

        if not fixed_fermi_level or self.fermi_levels is None:
            self.fermi_levels = np.array(fermi_levels) / Ha

        for f_n, kpt in zip(f_qn, self.kpt_u):
            kpt.f_n = f_n * (kpt.weightk * degeneracy)

        return e_entropy * degeneracy / Ha

    def set_positions(self, spos_ac, atom_partition=None):
        self.positions_set = False
        # rank_a = self.gd.get_ranks_from_positions(spos_ac)
        # atom_partition = AtomPartition(self.gd.comm, rank_a)
        # XXX pass AtomPartition around instead of spos_ac?
        # All the classes passing around spos_ac end up needing the ranks
        # anyway.

        if atom_partition is None:
            rank_a = self.gd.get_ranks_from_positions(spos_ac)
            atom_partition = AtomPartition(self.gd.comm, rank_a)

        if self.atom_partition is not None and self.kpt_u[0].P_ani is not None:
            with self.timer('Redistribute'):
                for kpt in self.kpt_u:
                    P = kpt.projections
                    assert self.atom_partition == P.atom_partition
                    kpt.projections = P.redist(atom_partition)
                    assert atom_partition == kpt.projections.atom_partition

        self.atom_partition = atom_partition
        self.kd.symmetry.check(spos_ac)
        self.spos_ac = spos_ac

    def allocate_arrays_for_projections(self, my_atom_indices):  # XXX unused
        if not self.positions_set and self.kpt_u[0]._projections is not None:
            # Projections have been read from file - don't delete them!
            pass
        else:
            nproj_a = [setup.ni for setup in self.setups]
            for kpt in self.kpt_u:
                kpt.projections = Projections(
                    self.bd.nbands, nproj_a,
                    self.atom_partition,
                    self.bd.comm,
                    collinear=self.collinear, spin=kpt.s, dtype=self.dtype)

    def collect_eigenvalues(self, k, s):
        return self.collect_array('eps_n', k, s)

    def collect_occupations(self, k, s):
        return self.collect_array('f_n', k, s)

    def collect_array(self, name, k, s, subset=None):
        """Helper method for collect_eigenvalues and collect_occupations.

        For the parallel case find the rank in kpt_comm that contains
        the (k,s) pair, for this rank, collect on the corresponding
        domain a full array on the domain master and send this to the
        global master."""

        kpt_qs = self.kpt_qs
        kpt_rank, q = self.kd.get_rank_and_index(k)
        if self.kd.comm.rank == kpt_rank:
            a_nx = getattr(kpt_qs[q][s], name)

            if subset is not None:
                a_nx = a_nx[subset]

            # Domain master send this to the global master
            if self.gd.comm.rank == 0:
                if self.bd.comm.size == 1:
                    if kpt_rank == 0:
                        return a_nx
                    else:
                        self.kd.comm.ssend(a_nx, 0, 1301)
                else:
                    b_nx = self.bd.collect(a_nx)
                    if self.bd.comm.rank == 0:
                        if kpt_rank == 0:
                            return b_nx
                        else:
                            self.kd.comm.ssend(b_nx, 0, 1301)

        elif self.world.rank == 0 and kpt_rank != 0:
            # Only used to determine shape and dtype of receiving buffer:
            a_nx = getattr(kpt_qs[0][0], name)

            if subset is not None:
                a_nx = a_nx[subset]

            b_nx = np.zeros((self.bd.nbands,) + a_nx.shape[1:],
                            dtype=a_nx.dtype)
            self.kd.comm.receive(b_nx, kpt_rank, 1301)
            return b_nx

        return np.zeros(0)  # see comment in get_wave_function_array() method

    def collect_auxiliary(self, value, k, s, shape=1, dtype=float):
        """Helper method for collecting band-independent scalars/arrays.

        For the parallel case find the rank in kpt_comm that contains
        the (k,s) pair, for this rank, collect on the corresponding
        domain a full array on the domain master and send this to the
        global master."""

        kpt_rank, q = self.kd.get_rank_and_index(k)

        if self.kd.comm.rank == kpt_rank:
            if isinstance(value, str):
                a_o = getattr(self.kpt_qs[q][s], value)
            else:
                u = q * self.nspins + s
                a_o = value[u]  # assumed list

            # Make sure data is a mutable object
            a_o = np.asarray(a_o)

            if a_o.dtype is not dtype:
                a_o = a_o.astype(dtype)

            # Domain master send this to the global master
            if self.gd.comm.rank == 0:
                if kpt_rank == 0:
                    return a_o
                else:
                    self.kd.comm.send(a_o, 0, 1302)

        elif self.world.rank == 0 and kpt_rank != 0:
            b_o = np.zeros(shape, dtype=dtype)
            self.kd.comm.receive(b_o, kpt_rank, 1302)
            return b_o

    def collect_projections(self, k, s):
        """Helper method for collecting projector overlaps across domains.

        For the parallel case find the rank in kpt_comm that contains
        the (k,s) pair, for this rank, send to the global master."""

        kpt_rank, q = self.kd.get_rank_and_index(k)

        if self.kd.comm.rank == kpt_rank:
            kpt = self.kpt_qs[q][s]
            P_nI = kpt.projections.collect()
            if self.world.rank == 0:
                return P_nI
            if P_nI is not None:
                self.kd.comm.send(np.ascontiguousarray(P_nI), 0, tag=117)
        if self.world.rank == 0:
            nproj = sum(setup.ni for setup in self.setups)
            if not self.collinear:
                nproj *= 2
            P_nI = np.empty((self.bd.nbands, nproj), self.dtype)
            self.kd.comm.receive(P_nI, kpt_rank, tag=117)
            return P_nI

    def get_wave_function_array(self, n, k, s, realspace=True, periodic=False):
        """Return pseudo-wave-function array on master.

        n: int
            Global band index.
        k: int
            Global IBZ k-point index.
        s: int
            Spin index (0 or 1).
        realspace: bool
            Transform plane wave or LCAO expansion coefficients to real-space.

        For the parallel case find the ranks in kd.comm and bd.comm
        that contains to (n, k, s), and collect on the corresponding
        domain a full array on the domain master and send this to the
        global master."""

        kpt_rank, q = self.kd.get_rank_and_index(k)
        band_rank, myn = self.bd.who_has(n)

        rank = self.world.rank

        if (self.kd.comm.rank == kpt_rank and
            self.bd.comm.rank == band_rank):
            u = q * self.nspins + s
            psit_G = self._get_wave_function_array(u, myn,
                                                   realspace, periodic)

            if realspace:
                psit_G = self.gd.collect(psit_G)

            if rank == 0:
                return psit_G

            # Domain master send this to the global master
            if self.gd.comm.rank == 0:
                psit_G = np.ascontiguousarray(psit_G)
                self.world.ssend(psit_G, 0, 1398)

        if rank == 0:
            # allocate full wave function and receive
            shape = () if self.collinear else (2,)
            psit_G = self.empty(shape, global_array=True,
                                realspace=realspace)
            # XXX this will fail when using non-standard nesting
            # of communicators.
            world_rank = (kpt_rank * self.gd.comm.size *
                          self.bd.comm.size +
                          band_rank * self.gd.comm.size)
            self.world.receive(psit_G, world_rank, 1398)
            return psit_G

        # We return a number instead of None on all the slaves.  Most of
        # the time the return value will be ignored on the slaves, but
        # in some cases it will be multiplied by some other number and
        # then ignored.  Allowing for this will simplify some code here
        # and there.
        return np.nan

    def get_homo_lumo(self, spin=None):
        """Return HOMO and LUMO eigenvalues."""
        if spin is None:
            if self.nspins == 1:
                return self.get_homo_lumo(0)
            h0, l0 = self.get_homo_lumo(0)
            h1, l1 = self.get_homo_lumo(1)
            return np.array([max(h0, h1), min(l0, l1)])

        n = self.nvalence // 2
        band_rank, myn = self.bd.who_has(n - 1)
        homo = -np.inf
        if self.bd.comm.rank == band_rank:
            for kpt in self.kpt_u:
                if kpt.s == spin:
                    homo = max(kpt.eps_n[myn], homo)
        homo = self.world.max(homo)

        lumo = np.inf
        if n < self.bd.nbands:  # there are not enough bands for LUMO
            band_rank, myn = self.bd.who_has(n)
            if self.bd.comm.rank == band_rank:
                for kpt in self.kpt_u:
                    if kpt.s == spin:
                        lumo = min(kpt.eps_n[myn], lumo)
            lumo = self.world.min(lumo)

        return np.array([homo, lumo])

    def write(self, writer):
        writer.write(version=2, ha=Ha)
        writer.write(fermi_levels=self.fermi_levels * Ha)
        writer.write(kpts=self.kd)
        self.write_projections(writer)
        self.write_eigenvalues(writer)
        self.write_occupations(writer)

    def write_projections(self, writer):
        nproj = sum(setup.ni for setup in self.setups)

        if self.collinear:
            shape = (self.nspins, self.kd.nibzkpts, self.bd.nbands, nproj)
        else:
            shape = (self.kd.nibzkpts, self.bd.nbands, 2, nproj)

        writer.add_array('projections', shape, self.dtype)

        for s in range(self.nspins):
            for k in range(self.kd.nibzkpts):
                P_nI = self.collect_projections(k, s)
                if not self.collinear and P_nI is not None:
                    P_nI.shape = (self.bd.nbands, 2, nproj)
                writer.fill(P_nI)

    def write_eigenvalues(self, writer):
        if self.collinear:
            shape = (self.nspins, self.kd.nibzkpts, self.bd.nbands)
        else:
            shape = (self.kd.nibzkpts, self.bd.nbands)

        writer.add_array('eigenvalues', shape)
        for s in range(self.nspins):
            for k in range(self.kd.nibzkpts):
                writer.fill(self.collect_eigenvalues(k, s) * Ha)

    def write_occupations(self, writer):

        if self.collinear:
            shape = (self.nspins, self.kd.nibzkpts, self.bd.nbands)
        else:
            shape = (self.kd.nibzkpts, self.bd.nbands)

        writer.add_array('occupations', shape)
        for s in range(self.nspins):
            for k in range(self.kd.nibzkpts):
                # Scale occupation numbers when writing:
                # XXX fix this in the code also ...
                weight = self.kd.weight_k[k] * 2 / self.nspins
                writer.fill(self.collect_occupations(k, s) / weight)

    def read(self, reader):
        r = reader.wave_functions
        # Backward compatibility:
        # Take parameters from main reader
        if 'ha' not in r:
            r.ha = reader.ha
        if 'version' not in r:
            r.version = reader.version

        if reader.version >= 3:
            self.fermi_levels = r.fermi_levels / r.ha
        else:
            o = reader.occupations
            self.fermi_levels = np.array(
                [o.fermilevel + o.split / 2,
                 o.fermilevel - o.split / 2]) / r.ha
            if self.occupations.name != 'fixmagmom':
                assert o.split == 0.0
                self.fermi_levels = self.fermi_levels[:1]

        if reader.version >= 2:
            kpts = r.kpts
            assert np.allclose(kpts.ibzkpts, self.kd.ibzk_kc)
            assert np.allclose(kpts.bzkpts, self.kd.bzk_kc)
            assert (kpts.bz2ibz == self.kd.bz2ibz_k).all()
            assert np.allclose(kpts.weights, self.kd.weight_k)

        self.read_projections(r)
        self.read_eigenvalues(r, r.version <= 0)
        self.read_occupations(r, r.version <= 0)

    def read_projections(self, reader):
        nslice = self.bd.get_slice()
        nproj_a = [setup.ni for setup in self.setups]
        atom_partition = AtomPartition(self.gd.comm,
                                       np.zeros(len(nproj_a), int))
        for u, kpt in enumerate(self.kpt_u):
            if self.collinear:
                index = (kpt.s, kpt.k)
            else:
                index = (kpt.k,)
            kpt.projections = Projections(
                self.bd.nbands, nproj_a,
                atom_partition, self.bd.comm,
                collinear=self.collinear, spin=kpt.s, dtype=self.dtype)
            if self.gd.comm.rank == 0:
                P_nI = reader.proxy('projections', *index)[nslice]
                if not self.collinear:
                    P_nI.shape = (self.bd.mynbands, -1)
                kpt.projections.matrix.array[:] = P_nI

    def read_eigenvalues(self, reader, old=False):
        nslice = self.bd.get_slice()
        for u, kpt in enumerate(self.kpt_u):
            if self.collinear:
                index = (kpt.s, kpt.k)
            else:
                index = (kpt.k,)
            eps_n = reader.proxy('eigenvalues', *index)[nslice]
            x = self.bd.mynbands - len(eps_n)  # missing bands?
            if x > 0:
                # Working on a real fix to this parallelization problem ...
                eps_n = np.pad(eps_n, (0, x), 'constant')
            if not old:  # skip for old tar-files gpw's
                eps_n /= reader.ha
            kpt.eps_n = eps_n

    def read_occupations(self, reader, old=False):
        nslice = self.bd.get_slice()
        for u, kpt in enumerate(self.kpt_u):
            if self.collinear:
                index = (kpt.s, kpt.k)
            else:
                index = (kpt.k,)
            f_n = reader.proxy('occupations', *index)[nslice]
            x = self.bd.mynbands - len(f_n)  # missing bands?
            if x > 0:
                # Working on a real fix to this parallelization problem ...
                f_n = np.pad(f_n, (0, x), 'constant')
            if not old:  # skip for old tar-files gpw's
                f_n *= kpt.weight
            kpt.f_n = f_n


def eigenvalue_string(wfs, comment=' '):
    """Write eigenvalues and occupation numbers into a string.

    The parameter comment can be used to comment out non-numers,
    for example to escape it for gnuplot.
    """
    tokens = []

    def add(*line):
        for token in line:
            tokens.append(token)
        tokens.append('\n')

    def eigs(k, s):
        eps_n = wfs.collect_eigenvalues(k, s)
        return eps_n * Ha

    def occs(k, s):
        occ_n = wfs.collect_occupations(k, s)
        return occ_n / wfs.kd.weight_k[k]

    if len(wfs.kd.ibzk_kc) == 1:
        if wfs.nspins == 1:
            add(comment, 'Band  Eigenvalues  Occupancy')
            eps_n = eigs(0, 0)
            f_n = occs(0, 0)
            if wfs.world.rank == 0:
                for n in range(wfs.bd.nbands):
                    add('%5d  %11.5f  %9.5f' % (n, eps_n[n], f_n[n]))
        else:
            add(comment, '                  Up                     Down')
            add(comment, 'Band  Eigenvalues  Occupancy  Eigenvalues  '
                'Occupancy')
            epsa_n = eigs(0, 0)
            epsb_n = eigs(0, 1)
            fa_n = occs(0, 0)
            fb_n = occs(0, 1)
            if wfs.world.rank == 0:
                for n in range(wfs.bd.nbands):
                    add('%5d  %11.5f  %9.5f  %11.5f  %9.5f' %
                        (n, epsa_n[n], fa_n[n], epsb_n[n], fb_n[n]))
        return ''.join(tokens)

    if len(wfs.kd.ibzk_kc) > 2:
        add('Showing only first 2 kpts')
        print_range = 2
    else:
        add('Showing all kpts')
        print_range = len(wfs.kd.ibzk_kc)

    if wfs.nvalence / 2. > 2:
        m = int(wfs.nvalence / 2. - 2)
    else:
        m = 0
    if wfs.bd.nbands - wfs.nvalence / 2. > 2:
        j = int(wfs.nvalence / 2. + 2)
    else:
        j = int(wfs.bd.nbands)

    if wfs.nspins == 1:
        add(comment, 'Kpt  Band  Eigenvalues  Occupancy')
        for i in range(print_range):
            eps_n = eigs(i, 0)
            f_n = occs(i, 0)
            if wfs.world.rank == 0:
                for n in range(m, j):
                    add('%3i %5d  %11.5f  %9.5f' % (i, n, eps_n[n], f_n[n]))
                add()
    else:
        add(comment, '                     Up                     Down')
        add(comment, 'Kpt  Band  Eigenvalues  Occupancy  Eigenvalues  '
            'Occupancy')

        for i in range(print_range):
            epsa_n = eigs(i, 0)
            epsb_n = eigs(i, 1)
            fa_n = occs(i, 0)
            fb_n = occs(i, 1)
            if wfs.world.rank == 0:
                for n in range(m, j):
                    add('%3i %5d  %11.5f  %9.5f  %11.5f  %9.5f' %
                        (i, n, epsa_n[n], fa_n[n], epsb_n[n], fb_n[n]))
                add()
    return ''.join(tokens)



class Contribution:
    def __init__(self, weight):
        self.weight = weight

    def get_name(self):
        raise NotImplementedError

    def get_desc(self):
        raise NotImplementedError

    def initialize(self, density, hamiltonian, wfs):
        self.density = density
        self.gd = density.gd
        self.finegd = density.finegd
        self.nspins = wfs.nspins
        self.wfs = wfs

    def initialize_1d(self, ae):
        self.ae = ae

    def initialize_from_atomic_orbitals(self, basis_functions):
        # Pass if contribution needs only density which is already initialized
        pass

    def calculate(self, e_g, n_sg, v_sg):
        raise NotImplementedError

    def calculate_energy_and_derivatives(self, setup, D_sp, H_sp):
        raise NotImplementedError

    def add_smooth_xc_potential_and_energy_1d(self, vt_g):
        raise NotImplementedError

    def get_extra_setup_data(self, extra_data):
        pass

    def write(self, writer):
        pass

    def read(self, reader):
        pass



"""Module for calculations using the Maximum Overlap Method (MOM).

See:

   https://arxiv.org/abs/2102.06542,
   :doi:`10.1021/acs.jctc.0c00597`.
"""

import numpy as np

from ase.units import Ha

from gpaw.occupations import FixedOccupationNumbers, ParallelLayout


def prepare_mom_calculation(calc,
                            atoms,
                            numbers,
                            use_projections=False,
                            update_numbers=True,
                            use_fixed_occupations=False,
                            width=0.0,
                            niter_width_update=40,
                            width_increment=0.0):
    """Helper function to prepare a calculator for a MOM calculation.

    calc: GPAW instance
        GPAW calculator object.
    atoms: ASE instance
        ASE atoms object.
    numbers: list (len=nspins) of lists (len=nbands)
        Occupation numbers (in the range from 0 to 1). Used
        for the initialization of the MOM reference orbitals.
    use_projections: bool
        If True, the occupied orbitals at iteration k are chosen
        as the orbitals ``|psi^(k)_m>`` with the biggest weights
        ``P_m`` evaluated as the projections onto the manifold of
        reference orbitals ``|psi_n>``: ``P_m = (Sum_n(|O_nm|^2))^0.5
        (O_nm = <psi_n|psi^(k)_m>)`` see
        :doi:`10.1021/acs.jctc.7b00994`. If False (default),
        the weights are evaluated as: ``P_m = max_n(|O_nm|)``
        see :doi:`10.1021/acs.jctc.0c00488`.
    update_numbers: bool
        If True (default), 'numbers' gets updated with the calculated
        occupation numbers, and when changing atomic positions
        the MOM reference orbitals will be initialized as the
        occupied orbitals found at convergence for the previous
        geometry. If False, when changing positions the MOM
        reference orbitals will be initialized as the orbitals
        of the previous geometry corresponding to the user-supplied
        'numbers'.
    use_fixed_occupations: bool
        If False (default), the MOM algorithm is used. If True,
        fixed occupations will be used.
    width: float
        Width of Gaussian function in eV for smearing of holes
        and excited electrons. The holes and excited electrons
        are found with respect to the zero-width ground-state
        occupations.
        See :doi:`10.1021/acs.jctc.0c00597`.
    niter_width_update: int
        Number of iterations after which the width of the
        Gaussian smearing function is increased.
    width_increment: float
        How much to increase the width of the Gaussian smearing
        function.
    """

    if calc.wfs is None:
        # We need the wfs object to initialize OccupationsMOM
        calc.initialize(atoms)

    occ_mom = OccupationsMOM(calc.wfs,
                             numbers,
                             use_projections,
                             update_numbers,
                             use_fixed_occupations,
                             width,
                             niter_width_update,
                             width_increment)
    calc.set(occupations=occ_mom)

    calc.log(occ_mom)


class OccupationsMOM:
    """MOM occupation class.

    The occupation numbers are found using a maximum overlap
    criterion and then broadcasted in occupations.py using the
    _calculate method of FixedOccupationNumbers.
    """

    def __init__(self,
                 wfs,
                 numbers,
                 use_projections=False,
                 update_numbers=True,
                 use_fixed_occupations=False,
                 width=0.0,
                 niter_width_update=10,
                 width_increment=0.0):
        self.wfs = wfs
        self.numbers = np.array(numbers)
        self.use_projections = use_projections
        self.update_numbers = update_numbers
        self.use_fixed_occupations = use_fixed_occupations
        self.width = width / Ha
        self.niter_width_update = niter_width_update
        self.width_increment = width_increment / Ha

        parallel_layout = ParallelLayout(self.wfs.bd,
                                         self.wfs.kd.comm,
                                         self.wfs.gd.comm)
        self.occ = FixedOccupationNumbers(numbers, parallel_layout)
        self.extrapolate_factor = self.occ.extrapolate_factor

        self.name = 'mom'
        self.iters = 0
        self.initialized = False

    def todict(self):
        dct = {'name': self.name,
               'numbers': self.numbers,
               'use_projections': self.use_projections,
               'update_numbers': self.update_numbers,
               'use_fixed_occupations': self.use_fixed_occupations}
        if self.width != 0.0:
            dct['width'] = self.width * Ha
            dct['niter_width_update'] = self.niter_width_update
            dct['width_increment'] = self.width_increment * Ha
        return dct

    def __str__(self):
        s = 'Excited-state calculation with Maximum Overlap Method\n'
        s += '  Gaussian smearing of holes and excited electrons: '
        if self.width == 0.0:
            s += 'off\n'
        else:
            s += '{0:.4f} eV\n'.format(self.width * Ha)
        return s

    def calculate(self,
                  nelectrons,
                  eigenvalues,
                  weights,
                  fermi_levels_guess):

        if not self.initialized:
            # If MOM reference orbitals are not initialized yet (e.g. when
            # the calculation is initialized from atomic densities), update
            # the occupation numbers according to the user-supplied 'numbers'
            self.occ.f_sn = self.numbers.copy()
            self.initialize_reference_orbitals()
        else:
            self.occ.f_sn = self.update_occupations()
            self.iters += 1

        f_qn, fermi_levels, e_entropy = self.occ.calculate(nelectrons,
                                                           eigenvalues,
                                                           weights,
                                                           fermi_levels_guess)

        return f_qn, fermi_levels, e_entropy

    def initialize_reference_orbitals(self):
        if self.wfs.kpt_u[0].f_n is None:
            # If the occupation numbers are not already available
            # (e.g. when the calculation is initialized from atomic
            # densities) we first need to take a step of eigensolver
            # and update the occupation numbers according to the
            # 'user-supplied' numbers before initializing the MOM
            # reference orbitals
            return

        self.iters = 0

        if self.use_fixed_occupations:
            self.initialized = True
            return

        # Initialize MOM reference orbitals for each equally
        # occupied subspace separately
        self.f_sn_unique = self.find_unique_occupation_numbers()
        if self.wfs.mode == 'lcao':
            self.c_ref = {}
            for kpt in self.wfs.kpt_u:
                self.c_ref[kpt.s] = {}
                for f_n_unique in self.f_sn_unique[kpt.s]:
                    occupied = self.f_sn_unique[kpt.s][f_n_unique]
                    self.c_ref[kpt.s][f_n_unique] = kpt.C_nM[occupied].copy()
        else:
            self.wf = {}
            self.p_an = {}
            for kpt in self.wfs.kpt_u:
                self.wf[kpt.s] = {}
                self.p_an[kpt.s] = {}
                for f_n_unique in self.f_sn_unique[kpt.s]:
                    occupied = self.f_sn_unique[kpt.s][f_n_unique]
                    # Pseudo wave functions
                    self.wf[kpt.s][f_n_unique] = kpt.psit_nG[occupied].copy()
                    # Atomic contributions times projector overlaps
                    self.p_an[kpt.s][f_n_unique] = \
                        dict([(a, np.dot(self.wfs.setups[a].dO_ii,
                                         P_ni[occupied].T))
                              for a, P_ni in kpt.P_ani.items()])

        self.initialized = True

    def update_occupations(self):
        if self.width != 0.0:
            if self.iters == 0:
                self.width_update_counter = 0
            if self.iters % self.niter_width_update == 0:
                self.gauss_width = self.width + self.width_update_counter \
                    * self.width_increment
                self.width_update_counter += 1

        if not self.use_fixed_occupations:
            f_sn = np.zeros_like(self.numbers)
            for kpt in self.wfs.kpt_u:
                # Compute weights with respect to each equally occupied
                # subspace of the reference orbitals and occupy orbitals
                # with biggest weights
                for f_n_unique in self.f_sn_unique[kpt.s]:
                    occupied = self.f_sn_unique[kpt.s][f_n_unique]
                    n_occ = len(f_sn[kpt.s][occupied])
                    unoccupied = f_sn[kpt.s] == 0

                    P = np.zeros(len(f_sn[kpt.s]))
                    P[unoccupied] = self.calculate_weights(kpt,
                                                           f_n_unique,
                                                           unoccupied)
                    P_max = np.argpartition(P, -n_occ)[-n_occ:]
                    f_sn[kpt.s][P_max] = f_n_unique

                if self.update_numbers:
                    self.numbers[kpt.s] = f_sn[kpt.s].copy()
        else:
            f_sn = self.numbers.copy()

        for kpt in self.wfs.kpt_u:
            if self.width != 0.0:
                orbs, f_sn_gs = self.find_hole_and_excited_orbitals(f_sn, kpt)
                if orbs:
                    for o in orbs:
                        mask, gauss = self.gaussian_smearing(kpt,
                                                             f_sn_gs,
                                                             o,
                                                             self.gauss_width)
                        f_sn_gs[mask] += (o[1] * gauss)
                    f_sn[kpt.s] = f_sn_gs.copy()

        return f_sn

    def calculate_weights(self, kpt, f_n_unique, unoccupied):
        if self.wfs.mode == 'lcao':
            O = np.dot(self.c_ref[kpt.s][f_n_unique].conj(),
                       np.dot(kpt.S_MM, kpt.C_nM[unoccupied].T))
        else:
            # Pseudo wave function overlaps
            O = self.wfs.integrate(self.wf[kpt.s][f_n_unique][:],
                                   kpt.psit_nG[unoccupied][:], True)

            # Atomic contributions
            O_corr = np.zeros_like(O)
            for a, p_a in self.p_an[kpt.s][f_n_unique].items():
                O_corr += np.dot(kpt.P_ani[a][unoccupied].conj(), p_a).T
            O_corr = np.ascontiguousarray(O_corr)
            self.wfs.gd.comm.sum(O_corr)

            # Sum pseudo wave and atomic contributions
            O += O_corr

        if self.use_projections:
            P = np.sum(abs(O) ** 2, axis=0)
            P = P ** 0.5
        else:
            P = np.amax(abs(O), axis=0)

        return P

    def find_hole_and_excited_orbitals(self, f_sn, kpt):
        # Zero-width occupations for ground state
        ne = int(f_sn[kpt.s].sum())
        f_sn_gs = np.zeros_like(f_sn[kpt.s])
        f_sn_gs[:ne] = 1.0
        f_sn_diff = f_sn[kpt.s] - f_sn_gs

        # Select hole and excited orbitals
        idxs = np.where(np.abs(f_sn_diff) > 1e-5)[0]
        w = f_sn_diff[np.abs(f_sn_diff) > 1e-5]
        orbs = list(zip(idxs, w))

        return orbs, f_sn_gs

    def gaussian_smearing(self, kpt, f_sn_gs, o, gauss_width):
        if o[1] < 0:
            mask = (f_sn_gs > 1e-8)
        else:
            mask = (f_sn_gs < 1e-8)

        e = kpt.eps_n[mask]
        de2 = -(e - kpt.eps_n[o[0]]) ** 2
        gauss = (1 / (gauss_width * np.sqrt(2 * np.pi)) *
                 np.exp(de2 / (2 * gauss_width ** 2)))
        gauss /= sum(gauss)

        return mask, gauss

    def find_unique_occupation_numbers(self):
        f_sn_unique = {}
        for kpt in self.wfs.kpt_u:
            f_sn_unique[kpt.s] = {}
            f_n = self.numbers[kpt.s]

            for f_n_unique in np.unique(f_n):
                if f_n_unique >= 1.0e-10:
                    f_sn_unique[kpt.s][f_n_unique] = f_n == f_n_unique

        return f_sn_unique



from math import pi, sqrt

import numpy as np
from ase.atoms import Atoms
from scipy.linalg import eigh

from gpaw.calculator import GPAW
from gpaw.wavefunctions.base import WaveFunctions
from gpaw.atom.radialgd import EquidistantRadialGridDescriptor
from gpaw.utilities import unpack
from gpaw.occupations import OccupationNumberCalculator
import gpaw.mpi as mpi


class MakeWaveFunctions:
    name = 'atompaw'
    interpolation = 9
    force_complex_dtype = False

    def __init__(self, gd):
        self.gd = gd

    def __call__(self, paw, gd, *args, **kwargs):
        return AtomWaveFunctions(self.gd, *args, **kwargs)


class AtomWaveFunctionsArray:
    def __init__(self, psit_nG):
        self.array = psit_nG


class AtomWaveFunctions(WaveFunctions):
    mode = 'atompaw'

    def initialize(self, density, hamiltonian, spos_ac):
        setup = self.setups[0]
        bf = AtomBasisFunctions(self.gd, setup.basis_functions_J)
        density.initialize_from_atomic_densities(bf)
        hamiltonian.update(density)
        return 0, 0

    def add_to_density_from_k_point(self, nt_sG, kpt):
        nt_sG[kpt.s] += np.dot(kpt.f_n / 4 / pi, kpt.psit_nG**2)

    def summary(self, log):
        log('Mode: Spherically symmetric atomic solver')


class AtomPoissonSolver:
    def get_description(self):
        return 'Radial equidistant'

    def set_grid_descriptor(self, gd):
        self.gd = gd
        self.relax_method = 0
        self.nn = 1

    def initialize(self):
        pass

    def get_stencil(self):
        return 'Exact'

    def solve(self, vHt_g, rhot_g, charge=0, timer=None):
        r = self.gd.r_g
        dp = rhot_g * r * self.gd.dr_g
        dq = dp * r
        p = np.add.accumulate(dp[::-1])[::-1]
        q = np.add.accumulate(dq[::-1])[::-1]
        vHt_g[:] = 4 * pi * (p - 0.5 * dp - (q - 0.5 * dq - q[0]) / r)
        return 1


class AtomEigensolver:
    def __init__(self, gd, f_sln):
        self.gd = gd
        self.f_sln = f_sln
        self.error = 0.0
        self.initialized = False

    def reset(self):
        self.initialized = False

    def initialize(self, wfs):
        r = self.gd.r_g
        h = r[0]
        N = len(r)
        lmax = len(self.f_sln[0]) - 1

        self.T_l = [np.eye(N) * (1.0 / h**2)]
        self.T_l[0].flat[1::N + 1] = -0.5 / h**2
        self.T_l[0].flat[N::N + 1] = -0.5 / h**2
        for l in range(1, lmax + 1):
            self.T_l.append(self.T_l[0] + np.diag(l * (l + 1) / 2.0 / r**2))

        self.S_l = [np.eye(N) for l in range(lmax + 1)]
        setup = wfs.setups[0]
        self.pt_j = np.array([[pt(x) * x**l for x in r]
                              for pt, l in zip(setup.pt_j, setup.l_j)])

        dS_ii = setup.dO_ii
        i1 = 0
        for pt1, l1 in zip(self.pt_j, setup.l_j):
            i2 = 0
            for pt2, l2 in zip(self.pt_j, setup.l_j):
                if l1 == l2 and l1 <= lmax:
                    self.S_l[l1] += (np.outer(pt1 * r, pt2 * r) *
                                     h * dS_ii[i1, i2])
                i2 += 2 * l2 + 1
            i1 += 2 * l1 + 1

        for kpt in wfs.kpt_u:
            kpt.eps_n = np.empty(wfs.bd.nbands)
            kpt.psit = AtomWaveFunctionsArray(self.gd.empty(wfs.bd.nbands))
            kpt.projections = {0: np.zeros((wfs.bd.nbands, len(dS_ii)))}

        self.initialized = True

    def iterate(self, hamiltonian, wfs):
        if not self.initialized:
            self.initialize(wfs)

        r = self.gd.r_g
        h = r[0]
        N = len(r)
        lmax = len(self.f_sln[0]) - 1
        setup = wfs.setups[0]

        e_n = np.zeros(N)

        for s in range(wfs.nspins):
            dH_ii = unpack(hamiltonian.dH_asp[0][s])
            kpt = wfs.kpt_u[s]
            N1 = 0
            for l in range(lmax + 1):
                H = self.T_l[l] + np.diag(hamiltonian.vt_sg[s])
                i1 = 0
                for pt1, l1 in zip(self.pt_j, setup.l_j):
                    i2 = 0
                    for pt2, l2 in zip(self.pt_j, setup.l_j):
                        if l1 == l2 == l:
                            H += (h * dH_ii[i1, i2] *
                                  np.outer(pt1 * r, pt2 * r))
                        i2 += 2 * l2 + 1
                    i1 += 2 * l1 + 1
                e_n, H = eigh(H, self.S_l[l].copy())

                for n in range(len(self.f_sln[s][l])):
                    N2 = N1 + 2 * l + 1
                    kpt.eps_n[N1:N2] = e_n[n]
                    kpt.psit_nG[N1:N2] = H[:, n] / r / sqrt(h)
                    i1 = 0
                    for pt, ll in zip(self.pt_j, setup.l_j):
                        i2 = i1 + 2 * ll + 1
                        if ll == l:
                            P = np.dot(kpt.psit_nG[N1], pt * r**2) * h
                            kpt.P_ani[0][N1:N2, i1:i2] = P * np.eye(2 * l + 1)
                        i1 = i2
                    N1 = N2


class AtomLocalizedFunctionsCollection:
    def __init__(self, gd, spline_aj):
        self.gd = gd
        spline = spline_aj[0][0]
        self.b_g = np.array([spline(r) for r in gd.r_g]) / sqrt(4 * pi)
        self.nfunctions = sum(2 * spline.get_angular_momentum_number() + 1
                              for spline in spline_aj[0])

    def set_positions(self, spos_ac, atom_partition=None):
        pass

    def add(self, a_xG, c_axi=1.0, q=-1):
        assert q == -1
        if isinstance(c_axi, float):
            a_xG += c_axi * self.b_g
        else:
            a_xG += c_axi[0][0] * self.b_g

    def integrate(self, a_g, c_ai, q=-1):
        assert a_g.ndim == 1
        assert q == -1
        c_ai[0][0] = self.gd.integrate(a_g, self.b_g)
        c_ai[0][1:] = 0.0

    def dict(self):
        return {0: np.empty(self.nfunctions)}


class AtomBasisFunctions:
    def __init__(self, gd, bfs_J):
        self.gd = gd
        self.bl_J = []
        self.Mmax = 0

        for bf in bfs_J:
            l = bf.get_angular_momentum_number()
            self.bl_J.append((np.array([bf(x) * x**l for x in gd.r_g]), l))
            self.Mmax += 2 * l + 1
        self.atom_indices = [0]
        self.my_atom_indices = [0]

    def set_positions(self, spos_ac):
        pass

    def add_to_density(self, nt_sG, f_asi):
        i = 0
        for b_g, l in self.bl_J:
            nt_sG += f_asi[0][:, i:i + 1] * (2 * l + 1) / 4 / pi * b_g**2
            i += 2 * l + 1


class AtomGridDescriptor(EquidistantRadialGridDescriptor):
    def __init__(self, h, rcut):
        ng = int(float(rcut) / h + 0.5) - 1
        rcut = ng * h
        EquidistantRadialGridDescriptor.__init__(self, h, ng, h0=h)
        self.sdisp_cd = np.empty((3, 2))
        self.comm = mpi.serial_comm
        self.pbc_c = np.zeros(3, bool)
        self.cell_cv = np.eye(3) * rcut
        self.N_c = np.ones(3, dtype=int) * 2 * ng
        self.h_cv = self.cell_cv / self.N_c
        self.dv = (rcut / 2 / ng)**3
        self.orthogonal = False
        self.parsize_c = (1, 1, 1)

    def get_ranks_from_positions(self, spos_ac):
        return np.array([0])

    def refine(self):
        return self

    def get_lfc(self, gd, spline_aj):
        return AtomLocalizedFunctionsCollection(gd, spline_aj)

    def integrate(self, a_xg, b_xg=None, global_integral=True):
        """Integrate function(s) in array over domain."""
        if b_xg is None:
            return np.dot(a_xg, self.dv_g)
        else:
            return np.dot(a_xg * b_xg, self.dv_g)

    def calculate_dipole_moment(self, rhot_g):
        return np.zeros(3)

    def symmetrize(self, a_g, op_scc, ft_sc=None):
        pass

    def get_grid_spacings(self):
        return self.h_cv.diagonal()

    def get_size_of_global_array(self, pad=None):
        return np.array(len(self.N_c))

    def new_descriptor(self, *args, **kwargs):
        return self

    def get_processor_position_from_rank(self, rank):
        return (0, 0, 0)


class AtomOccupations(OccupationNumberCalculator):
    extrapolate_factor = 0.0

    def __init__(self, f_sln):
        self.f_sln = f_sln
        OccupationNumberCalculator.__init__(self)

    def _calculate(self,
                   nelectrons,
                   eig_qn,
                   weight_q,
                   f_qn,
                   fermi_level_guess):
        for s, f_n in enumerate(f_qn):
            n1 = 0
            for l, f0_n in enumerate(self.f_sln[s]):
                for f in f0_n:
                    n2 = n1 + 2 * l + 1
                    f_n[n1:n2] = f / (2 * l + 1) / 2
                    n1 = n2

        return np.inf, 0.0


class AtomPAW(GPAW):
    def __init__(self, symbol, f_sln, h=0.05, rcut=10.0, **kwargs):
        assert len(f_sln) in [1, 2]
        self.symbol = symbol

        gd = AtomGridDescriptor(h, rcut)
        GPAW.__init__(self,
                      mode=MakeWaveFunctions(gd),
                      eigensolver=AtomEigensolver(gd, f_sln),
                      poissonsolver=AtomPoissonSolver(),
                      nbands=sum([(2 * l + 1) * len(f_n)
                                  for l, f_n in enumerate(f_sln[0])]),
                      communicator=mpi.serial_comm,
                      parallel=dict(augment_grids=False),
                      occupations=AtomOccupations(f_sln),
                      **kwargs)
        # Initialize function will raise an error unless we set a (bogus) cell
        self.initialize(Atoms(symbol, calculator=self,
                              cell=np.eye(3)))
        self.density.charge_eps = 1e-3
        self.calculate(system_changes=['positions'])

    def dry_run(self):
        pass

    def state_iter(self):
        """Yield the tuples (l, n, f, eps, psit_G) of states.

        Skips degenerate states."""
        f_sln = self.wfs.occupations.f_sln
        assert len(f_sln) == 1, 'Not yet implemented with more spins'
        f_ln = f_sln[0]
        kpt = self.wfs.kpt_u[0]

        band = 0
        for l, f_n in enumerate(f_ln):
            for n, f in enumerate(f_n):
                psit_G = kpt.psit_nG[band]
                eps = kpt.eps_n[band]
                yield l, n, f, eps, psit_G
                band += 2 * l + 1

    def extract_basis_functions(self, basis_name='atompaw.sz'):
        """Create BasisFunctions object with pseudo wave functions."""
        from gpaw.basis_data import Basis, BasisFunction
        assert self.wfs.nspins == 1

        d = self.wfs.gd.r_g[0]
        ng = self.wfs.gd.N + 1
        rgd = EquidistantRadialGridDescriptor(d, ng)
        basis = Basis(self.symbol, basis_name, readxml=False, rgd=rgd)
        basis.generatorattrs = {}  # attrs of the setup maybe
        basis.generatordata = 'AtomPAW'  # version info too?

        bf_j = basis.bf_j
        for l, n, f, eps, psit_G in self.state_iter():
            bf_g = rgd.empty()
            bf_g[0] = 0.0
            bf_g[1:] = psit_G
            bf_g *= np.sign(psit_G[-1])

            # If there's no node at zero, we shouldn't set bf_g to zero
            # We'll make an ugly hack
            if abs(bf_g[1]) > 3.0 * abs(bf_g[2] - bf_g[1]):
                bf_g[0] = bf_g[1]
            bf = BasisFunction(n, l, self.wfs.gd.r_g[-1], bf_g,
                               '%s%d e=%.3f f=%.3f' % ('spdfgh'[l], n, eps, f))
            bf_j.append(bf)
        return basis



"""Check for tunability of gamma for yukawa potential."""
import pytest
from ase import Atoms
from gpaw import GPAW
from gpaw.cluster import Cluster
from gpaw.eigensolvers import RMMDIIS
from gpaw.xc.hybrid import HybridXC
from gpaw.occupations import FermiDirac
from gpaw.test import gen
import _gpaw


@pytest.mark.hybrids
def test_rsf_yukawa_rsf_general(in_tmp_dir, add_cwd_to_setup_paths):
    libxc_version = getattr(_gpaw, 'libxc_version', '2.x.y')
    if int(libxc_version.split('.')[0]) < 3:
        from unittest import SkipTest
        raise SkipTest

    for atom in ['Be']:
        gen(atom, xcname='PBE', scalarrel=True, exx=True,
            yukawa_gamma=0.83, gpernode=149)

    h = 0.35
    be = Cluster(Atoms('Be', positions=[[0, 0, 0]]))
    be.minimal_box(3.0, h=h)

    c = {'energy': 0.05, 'eigenstates': 0.05, 'density': 0.05}

    IP = 8.79

    xc = HybridXC('LCY-PBE', omega=0.83)
    fname = 'Be_rsf.gpw'

    calc = GPAW(txt='Be.txt', xc=xc, convergence=c,
                eigensolver=RMMDIIS(), h=h,
                occupations=FermiDirac(width=0.0), spinpol=False)
    be.calc = calc
    energy_083 = be.get_potential_energy()
    (eps_homo, eps_lumo) = calc.get_homo_lumo()
    assert eps_homo == pytest.approx(-IP, abs=0.15)
    xc2 = 'LCY-PBE'
    energy_075 = calc.get_xc_difference(HybridXC(xc2))
    assert energy_083 - energy_075 == pytest.approx(21.13, abs=0.2)
    calc.write(fname)
    calc2 = GPAW(fname)
    xc = calc2.hamiltonian.xc
    assert xc.name == 'LCY-PBE', 'wrong name for functional'
    assert xc.omega == 0.83, 'wrong value for RSF omega'



from __future__ import annotations

from math import pi
from types import SimpleNamespace

import numpy as np
from ase.data import atomic_numbers, covalent_radii
from ase.neighborlist import neighbor_list
from ase.units import Bohr, Ha

from gpaw.core.arrays import DistributedArrays
from gpaw.core.atom_arrays import AtomArraysLayout
from gpaw.core.domain import Domain
from gpaw.core.matrix import Matrix
from gpaw.lcao.tci import TCIExpansions
from gpaw.lfc import BasisFunctions
from gpaw.mpi import MPIComm, serial_comm
from gpaw.new import zip
from gpaw.new.calculation import DFTState
from gpaw.new.lcao.builder import LCAODFTComponentsBuilder, create_lcao_ibzwfs
from gpaw.new.lcao.hamiltonian import CollinearHamiltonianMatrixCalculator
from gpaw.new.lcao.wave_functions import LCAOWaveFunctions
from gpaw.new.pot_calc import PotentialCalculator
from gpaw.setup import Setup
from gpaw.spline import Spline
from gpaw.utilities.timing import NullTimer
from gpaw.typing import Array3D


class TBHamiltonianMatrixCalculator(CollinearHamiltonianMatrixCalculator):
    def _calculate_potential_matrix(self,
                                    wfs: LCAOWaveFunctions,
                                    V_xMM: Array3D = None) -> Matrix:
        return wfs.V_MM


class TBHamiltonian:
    def __init__(self,
                 basis: BasisFunctions):
        self.basis = basis

    def apply(self):
        raise NotImplementedError

    def create_hamiltonian_matrix_calculator(
            self,
            state: DFTState) -> TBHamiltonianMatrixCalculator:
        dH_saii = [{a: dH_sii[s]
                    for a, dH_sii in state.potential.dH_asii.items()}
                   for s in range(state.density.ncomponents)]

        V_sxMM = [np.zeros(0) for _ in range(state.density.ncomponents)]

        return TBHamiltonianMatrixCalculator(V_sxMM, dH_saii, self.basis)


class NoGrid(Domain):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._gd = SimpleNamespace(
            get_grid_spacings=lambda: [0, 0, 0],
            cell_cv=self.cell_cv,
            pbc_c=self.pbc_c,
            N_c=[0, 0, 0],
            dv=0.0)
        self.size = (0, 0, 0)

    def empty(self, shape=(), comm=serial_comm, xp=None):
        return DummyFunctions(self, shape, comm)

    def ranks_from_fractional_positions(self, fracpos_ac):
        return np.zeros(len(fracpos_ac), int)


class DummyFunctions(DistributedArrays[NoGrid]):
    def __init__(self,
                 grid: NoGrid,
                 dims: int | tuple[int, ...] = (),
                 comm: MPIComm = serial_comm):
        DistributedArrays. __init__(self, dims, (),
                                    comm, grid.comm, None, np.nan,
                                    grid.dtype)
        self.desc = grid

    def integrate(self):
        return np.ones(self.dims)

    def new(self):
        return self

    def __getitem__(self, index):
        return DummyFunctions(self.desc, comm=self.comm)

    def moment(self):
        return np.zeros(3)

    def to_xp(self, xp):
        return self


class PSCoreDensities:
    def __init__(self, grid, fracpos_ac):
        self.layout = AtomArraysLayout([1] * len(fracpos_ac),
                                       grid.comm)

    def to_uniform_grid(self, out, scale):
        pass


class TBPotentialCalculator(PotentialCalculator):
    def __init__(self,
                 xc,
                 setups,
                 nct_R,
                 atoms):
        super().__init__(xc, None, setups, nct_R,
                         atoms.get_scaled_positions())
        self.atoms = atoms.copy()
        self.force_av = None
        self.stress_vv = None

    def calculate_charges(self, vHt_r):
        return AtomArraysLayout(
            [9] * len(self.atoms),
            self.nct_R.comm).zeros()

    def calculate_pseudo_potential(self, density, vHt_r):
        vt_sR = density.nt_sR

        atoms = self.atoms
        energy, force_av, stress_vv = pairpot(atoms)
        energy /= Ha
        self.force_av = force_av * Bohr / Ha

        vol = abs(np.linalg.det(atoms.cell[atoms.pbc][:, atoms.pbc]))
        self.stress_vv = stress_vv / vol * Bohr**atoms.pbc.sum() / Ha

        return {'kinetic': 0.0,
                'coulomb': 0.0,
                'zero': 0.0,
                'xc': energy,
                'external': 0.0}, vt_sR, vHt_r

    def _move(self, fracpos_ac, ndensities):
        self.atoms.set_scaled_positions(fracpos_ac)
        self.force_av = None
        self.stress_vv = None

    def force_contributions(self, state):
        return {}, {}, {a: self.force_av[a:a + 1]
                        for a in state.density.D_asii.keys()}

    def stress_contribution(self, state):
        return self.stress_vv


class DummyXC:
    no_forces = False
    xc = None

    def calculate_paw_correction(self, setup, D_sp, dH_sp):
        return 0.0


class TBSCFLoop:
    def __init__(self, hamiltonian, occ_calc, eigensolver, world):
        self.hamiltonian = hamiltonian
        self.occ_calc = occ_calc
        self.eigensolver = eigensolver
        self.world = world

    def iterate(self,
                state,
                pot_calc,
                convergence=None,
                maxiter=None,
                calculate_forces=None,
                log=None):
        self.eigensolver.iterate(state, self.hamiltonian)
        state.ibzwfs.calculate_occs(self.occ_calc)
        yield
        state.potential, state.vHt_x, _ = pot_calc.calculate(
            state.density, state.vHt_x)


class DummyBasis:
    def __init__(self, natoms):
        self.my_atom_indices = np.arange(natoms)

    def add_to_density(self, nt_sR, f_asi):
        pass

    def construct_density(self, rho_MM, nt_G, q):
        pass


class TBDFTComponentsBuilder(LCAODFTComponentsBuilder):
    def check_cell(self, cell):
        pass

    def create_uniform_grids(self):
        grid = NoGrid(
            self.atoms.cell.complete() / Bohr,
            self.atoms.pbc,
            dtype=self.dtype,
            comm=self.communicators['d'])
        return grid, grid

    def get_pseudo_core_densities(self):
        return PSCoreDensities(self.grid, self.fracpos_ac)

    def create_basis_set(self):
        self.basis = DummyBasis(len(self.atoms))
        return self.basis

    def create_hamiltonian_operator(self):
        return TBHamiltonian(self.basis)

    def create_potential_calculator(self):
        xc = DummyXC()
        return TBPotentialCalculator(xc, self.setups, self.nct_R, self.atoms)

    def create_scf_loop(self):
        occ_calc = self.create_occupation_number_calculator()
        hamiltonian = self.create_hamiltonian_operator()
        eigensolver = self.create_eigensolver(hamiltonian)
        return TBSCFLoop(hamiltonian, occ_calc, eigensolver,
                         self.communicators['w'])

    def create_ibz_wave_functions(self,
                                  basis: BasisFunctions,
                                  potential,
                                  *,
                                  log=None,
                                  coefficients=None):
        assert self.communicators['w'].size == 1

        ibzwfs, tciexpansions = create_lcao_ibzwfs(
            basis, potential,
            self.ibz, self.communicators, self.setups,
            self.fracpos_ac, self.grid, self.dtype,
            self.nbands, self.ncomponents, self.atomdist, self.nelectrons)

        vtphit: dict[Setup, list[Spline]] = {}

        for setup in self.setups.setups.values():
            try:
                vt_r = setup.vt_g
            except AttributeError:
                vt_r = calculate_pseudo_potential(setup, self.xc.xc)[0]

            vt_r[-1] = 0.0  # ???
            vt = setup.rgd.spline(vt_r, points=300)
            vtphit_j = []
            for phit in setup.basis_functions_J:
                rc = phit.get_cutoff()
                r_g = np.linspace(0, rc, 150)
                vt_g = vt.map(r_g) / (4 * pi)**0.5
                phit_g = phit.map(r_g)
                vtphit_j.append(Spline(phit.l, rc, vt_g * phit_g))
            vtphit[setup] = vtphit_j

        vtciexpansions = TCIExpansions([s.basis_functions_J
                                        for s in self.setups],
                                       [vtphit[s] for s in self.setups],
                                       tciexpansions.I_a)

        kpt_qc = np.array([wfs.kpt_c for wfs in ibzwfs])
        manytci = vtciexpansions.get_manytci_calculator(
            self.setups, self.grid._gd, self.fracpos_ac,
            kpt_qc, self.dtype, NullTimer())

        manytci.Pindices = manytci.Mindices
        my_atom_indices = basis.my_atom_indices

        for wfs, V_MM in zip(ibzwfs, manytci.P_qIM(my_atom_indices)):
            V_MM = V_MM.toarray()
            V_MM += V_MM.T.conj().copy()
            M1 = 0
            for m in manytci.Mindices.nm_a:
                M2 = M1 + m
                V_MM[M1:M2, M1:M2] *= 0.5
                M1 = M2
            wfs.V_MM = Matrix(M2, M2, data=V_MM)

        return ibzwfs


def pairpot(atoms):
    """Simple pair-potential for testing.

    >>> from ase import Atoms
    >>> r = covalent_radii[1]
    >>> atoms = Atoms('H2', [(0, 0, 0), (0, 0, 2 * r)])
    >>> e, f, s = pairpot(atoms)
    >>> print(f'{e:.6f} eV')
    -9.677419 eV
    >>> f
    array([[0., 0., 0.],
           [0., 0., 0.]])

    """
    radii = {}
    symbol_a = atoms.symbols
    for symbol in symbol_a:
        radii[symbol] = covalent_radii[atomic_numbers[symbol]]

    r0 = {}
    for s1, r1 in radii.items():
        for s2, r2 in radii.items():
            r0[(s1, s2)] = r1 + r2
    rcutmax = 2 * max(r0.values(), default=1.0)

    energy = 0.0
    force_av = np.zeros((len(atoms), 3))
    stress_vv = np.zeros((3, 3))

    for i, j, d, D_v in zip(*neighbor_list('ijdD', atoms, rcutmax)):
        d0 = r0[(symbol_a[i], symbol_a[j])]
        e0 = 6.0 / d0
        x = d0 / d
        if x > 0.5:
            energy += 0.5 * e0 * (-5 + x * (24 + x * (-36 + 16 * x)))
            f = -0.5 * e0 * (24 + x * (-72 + 48 * x)) * d0 / d**2
            F_v = D_v * f / d
            force_av[i] += F_v
            force_av[j] -= F_v
            # print(i, j, d, D_v, F_v)
            stress_vv += np.outer(F_v, D_v)

    return energy, force_av, stress_vv


def calculate_pseudo_potential(setup: Setup, xc):
    phit_jg = np.array(setup.data.phit_jg)
    rgd = setup.rgd

    # Density:
    nt_g = np.einsum('jg, j, jg -> g',
                     phit_jg, setup.f_j, phit_jg) / (4 * pi)
    nt_g += setup.data.nct_g * (1 / (4 * pi)**0.5)

    # XC:
    vt_g = rgd.zeros()
    xc.calculate_spherical(rgd, nt_g[np.newaxis], vt_g[np.newaxis])

    # Zero-potential:
    vt_g += setup.data.vbar_g / (4 * pi)**0.5

    # Coulomb:
    g_g = setup.ghat_l[0].map(rgd.r_g)
    Q = -rgd.integrate(nt_g) / rgd.integrate(g_g)
    rhot_g = nt_g + Q * g_g
    vHtr_g = rgd.poisson(rhot_g)

    W = rgd.integrate(g_g * vHtr_g, n=-1) / (4 * pi)**0.5

    vtr_g = vt_g * rgd.r_g + vHtr_g

    vtr_g[1:] /= rgd.r_g[1:]
    vtr_g[0] = vtr_g[1]

    return vtr_g * (4 * pi)**0.5, W


def poly():
    """Polynomium used for pair potential."""
    import matplotlib.pyplot as plt
    c = np.linalg.solve([[1, 0.5, 0.25, 0.125],
                         [1, 1, 1, 1],
                         [0, 1, 1, 0.75],
                         [0, 1, 2, 3]],
                        [0, -1, 0, 0])
    print(c)
    d = np.linspace(0.5, 2, 101)
    plt.plot(d, c[0] + c[1] / d + c[2] / d**2 + c[3] / d**3)
    plt.show()


if __name__ == '__main__':
    poly()



import hashlib
import os
import re
import xml.sax
from glob import glob
from math import pi, sqrt
from pathlib import Path
from typing import IO, Tuple

import numpy as np
from ase.data import atomic_names, atomic_numbers
from ase.units import Bohr, Ha

from gpaw import setup_paths
from gpaw.atom.radialgd import (AbinitRadialGridDescriptor,
                                AERadialGridDescriptor)
from gpaw.atom.shapefunc import shape_functions
from gpaw.mpi import broadcast
from gpaw.xc.pawcorrection import PAWXCCorrection


class SetupData:
    """Container class for persistent setup attributes and XML I/O."""
    def __init__(self, symbol, xcsetupname,
                 name='paw', readxml=True,
                 zero_reference=False, world=None,
                 generator_version=None):
        self.symbol = symbol
        self.setupname = xcsetupname
        self.name = name
        self.zero_reference = zero_reference
        self.generator_version = generator_version

        # Default filename if this setup is written
        if name is None or name == 'paw':
            self.stdfilename = f'{symbol}.{self.setupname}'
        else:
            self.stdfilename = f'{symbol}.{name}.{self.setupname}'

        self.filename = None  # full path if this setup was loaded from file
        self.fingerprint = None  # hash value of file data if applicable

        self.Z = None
        self.Nc = None
        self.Nv = None

        # Quantum numbers, energies
        self.n_j = []
        self.l_j = []
        self.l_orb_J = self.l_j  # pointer to same list!
        self.f_j = []
        self.eps_j = []
        self.e_kin_jj = None  # <phi | T | phi> - <phit | T | phit>

        self.rgd = None

        # Parameters for compensation charge expansion functions:
        self.shape_function = {'type': 'undefined', 'rc': np.nan}

        # State identifier, like "X-2s" or "X-p1", where X is chemical symbol,
        # for bound and unbound states
        self.id_j = []

        # Partial waves, projectors
        self.phi_jg = []
        self.phit_jg = []
        self.pt_jg = []
        self.rcut_j = []

        # Densities, potentials
        self.nc_g = None
        self.nct_g = None
        self.nvt_g = None
        self.vbar_g = None
        self.vt_g = None

        # Kinetic energy densities of core electrons
        self.tauc_g = None
        self.tauct_g = None

        # Reference energies
        self.e_kinetic = 0.0
        self.e_xc = 0.0
        self.e_electrostatic = 0.0
        self.e_total = 0.0
        self.e_kinetic_core = 0.0

        # Generator may store description of setup in this string
        self.generatorattrs = []
        self.generatordata = ''

        # Optional quantities, normally not used
        self.X_p = None
        self.X_wp = {}
        self.X_pg = None
        self.ExxC = None
        self.ExxC_w = {}
        self.X_gamma = None
        self.extra_xc_data = {}
        self.phicorehole_g = None
        self.fcorehole = 0.0
        self.lcorehole = None
        self.ncorehole = None
        self.core_hole_e = None
        self.core_hole_e_kin = None
        self.has_corehole = False

        # Parameters for zero-potential:
        self.l0 = None
        self.e0 = None
        self.r0 = None
        self.nderiv0 = None

        self.orbital_free = False  # orbital-free DFT

        self.version = None

        if readxml:
            self.read_xml(world=world)

    def __repr__(self):
        return ('{0}({symbol!r}, {setupname!r}, name={name!r}, '
                'generator_version={generator_version!r}, ...)'
                .format(self.__class__.__name__, **vars(self)))

    def append(self, n, l, f, e, rcut, phi_g, phit_g, pt_g):
        self.n_j.append(n)
        self.l_j.append(l)
        self.f_j.append(f)
        self.eps_j.append(e)
        self.rcut_j.append(rcut)
        self.phi_jg.append(phi_g)
        self.phit_jg.append(phit_g)
        self.pt_jg.append(pt_g)

    def read_xml(self, source=None, world=None):
        PAWXMLParser(self).parse(source=source, world=world)
        nj = len(self.l_j)
        self.e_kin_jj.shape = (nj, nj)

    def is_compatible(self, xc):
        return xc.get_setup_name() == self.setupname

    def print_info(self, text, setup):
        if self.phicorehole_g is None:
            text(self.symbol + ':')
        else:
            text(f'{self.symbol}:  # ({self.fcorehole:.1f} core hole)')
        text('  name:', atomic_names[atomic_numbers[self.symbol]])
        text('  id:', self.fingerprint)
        text('  Z:', self.Z)
        text('  valence:', self.Nv)
        if self.phicorehole_g is None:
            text('  core: %d' % self.Nc)
        else:
            text(f'  core: {self.Nc:.1f}')
        text('  charge:', self.Z - self.Nv - self.Nc)
        if setup.hubbard_u is not None:
            description = ''.join([f'  {line}' for line
                                   in setup.hubbard_u.descriptions()])
            text(description)
        text('  file:', self.filename)
        sf = self.shape_function
        text(f'  compensation charges: {{type: {sf["type"]},\n'
             f'                         rc: {sf["rc"] * Bohr:.2f},\n'
             f'                         lmax: {setup.lmax}}}')
        text(f'  cutoffs: {{filter: {setup.rcutfilter * Bohr:.2f},\n'
             f'            core: {setup.rcore * Bohr:.2f}}}')
        text('  valence states:')
        text('    #              energy  rcut')
        j = 0
        for n, l, f, eps in zip(self.n_j, self.l_j, self.f_j, self.eps_j):
            if n > 0:
                f = f'({f:.2f})'
                text('    - %d%s%-5s %9.3f   %5.3f' % (
                    n, 'spdf'[l], f, eps * Ha, self.rcut_j[j] * Bohr))
            else:
                text('    -  %s       %9.3f   %5.3f' % (
                    'spdf'[l], eps * Ha, self.rcut_j[j] * Bohr))
            j += 1
        text()

    def create_compensation_charge_functions(self, lmax):
        """Create shape functions used to expand compensation charges."""
        g_lg = shape_functions(self.rgd, **self.shape_function, lmax=lmax)
        return g_lg

    def get_smooth_core_density_integral(self, Delta0):
        return -Delta0 * sqrt(4 * pi) - self.Z + self.Nc

    def get_overlap_correction(self, Delta0_ii):
        return sqrt(4.0 * pi) * Delta0_ii

    def get_linear_kinetic_correction(self, T0_qp):
        e_kin_jj = self.e_kin_jj
        nj = len(e_kin_jj)
        K_q = []
        for j1 in range(nj):
            for j2 in range(j1, nj):
                K_q.append(e_kin_jj[j1, j2])
        K_p = sqrt(4 * pi) * np.dot(K_q, T0_qp)
        return K_p

    def find_core_density_cutoff(self, nc_g):
        if self.Nc == 0:
            return 1.0
        else:
            rgd = self.rgd
            N = 0.0
            g = self.rgd.N - 1
            while N < 1e-7:
                N += sqrt(4 * pi) * nc_g[g] * rgd.r_g[g]**2 * rgd.dr_g[g]
                g -= 1
            return rgd.r_g[g]

    def get_xc_correction(self, rgd, xc, gcut2, lcut):
        phicorehole_g = self.phicorehole_g
        if phicorehole_g is not None:
            phicorehole_g = phicorehole_g[:gcut2].copy()

        xc_correction = PAWXCCorrection(
            [phi_g[:gcut2] for phi_g in self.phi_jg],
            [phit_g[:gcut2] for phit_g in self.phit_jg],
            self.nc_g[:gcut2] / sqrt(4 * pi),
            self.nct_g[:gcut2] / sqrt(4 * pi),
            rgd,
            list(enumerate(self.l_j)),
            min(2 * lcut, 4),
            self.e_xc,
            phicorehole_g,
            self.fcorehole,
            None if self.tauc_g is None else self.tauc_g[:gcut2].copy(),
            None if self.tauct_g is None else self.tauct_g[:gcut2].copy())

        return xc_correction

    def write_xml(self) -> None:
        with open(self.stdfilename, 'w') as fd:
            self._write_xml(fd)

    def _write_xml(self, xml: IO[str]) -> None:
        l_j = self.l_j

        print('<?xml version="1.0"?>', file=xml)
        print(f'<paw_dataset version="{self.version}">',
              file=xml)
        name = atomic_names[atomic_numbers[self.symbol]].title()
        comment1 = name + ' setup for the Projector Augmented Wave method.'
        comment2 = 'Units: Hartree and Bohr radii.'
        comment2 += ' ' * (len(comment1) - len(comment2))
        print('  <!--', comment1, '-->', file=xml)
        print('  <!--', comment2, '-->', file=xml)

        print(('  <atom symbol="%s" Z="%r" core="%r" valence="%r"/>' %
               (self.symbol, self.Z, self.Nc, self.Nv)), file=xml)
        if self.orbital_free:
            type = 'OFDFT'
            name = self.setupname
        elif self.setupname == 'LDA':
            type = 'LDA'
            name = 'PW'
        else:
            type = 'GGA'
            name = self.setupname
        print(f'  <xc_functional type="{type}" name="{name}"/>',
              file=xml)
        gen_attrs = ' '.join([f'{key}="{value}"' for key, value
                              in self.generatorattrs])
        print(f'  <generator {gen_attrs}>', file=xml)
        print(f'    {self.generatordata}', file=xml)
        print('  </generator>', file=xml)
        print(f'  <ae_energy kinetic="{self.e_kinetic!r}" xc="{self.e_xc!r}"',
              file=xml)
        print('             electrostatic="%r" total="%r"/>' %
              (self.e_electrostatic, self.e_total), file=xml)

        print(f'  <core_energy kinetic="{self.e_kinetic_core!r}"/>', file=xml)
        print('  <valence_states>', file=xml)
        line1 = '    <state n="%d" l="%d" f="%r" rc="%r" e="%r" id="%s"/>'
        line2 = '    <state       l="%d"        rc="%r" e="%r" id="%s"/>'

        for id, l, n, f, e, rc in zip(self.id_j, l_j, self.n_j, self.f_j,
                                      self.eps_j, self.rcut_j):
            if n > 0:
                print(line1 % (n, l, f, rc, e, id), file=xml)
            else:
                print(line2 % (l, rc, e, id), file=xml)
        print('  </valence_states>', file=xml)

        print(self.rgd.xml('g1'), file=xml)

        print('  <shape_function type="{type}" rc="{rc}"/>'
              .format(**self.shape_function), file=xml)

        if self.r0 is None:
            # Old setups:
            xml.write('  <zero_potential grid="g1">\n')
        elif self.l0 is None:
            xml.write('  <zero_potential type="polynomial" ' +
                      'nderiv="%d" r0="%r" grid="g1">\n' %
                      (self.nderiv0, self.r0))
        else:
            xml.write(('  <zero_potential type="%s" ' +
                       'e0="%r" nderiv="%d" r0="%r" grid="g1">\n') %
                      ('spdfg'[self.l0], self.e0, self.nderiv0, self.r0))

        for x in self.vbar_g:
            print(f'{x!r}', end=' ', file=xml)
        print('\n  </zero_potential>', file=xml)

        if self.has_corehole:
            print((('  <core_hole_state state="%d%s" ' +
                    'removed="%r" eig="%r" ekin="%r">') %
                   (self.ncorehole, 'spdf'[self.lcorehole],
                    self.fcorehole,
                    self.core_hole_e, self.core_hole_e_kin)), file=xml)
            for x in self.phicorehole_g:
                print(f'{x!r}', end=' ', file=xml)
            print('\n  </core_hole_state>', file=xml)

        for name, a in [('ae_core_density', self.nc_g),
                        ('pseudo_core_density', self.nct_g),
                        ('ae_core_kinetic_energy_density', self.tauc_g),
                        ('pseudo_core_kinetic_energy_density', self.tauct_g)]:
            print(f'  <{name} grid="g1">\n    ', end=' ', file=xml)
            for x in a:
                print(f'{x!r}', end=' ', file=xml)
            print(f'\n  </{name}>', file=xml)

        # Print xc-specific data to setup file (used so for KLI and GLLB)
        for name, a in self.extra_xc_data.items():
            newname = 'GLLB_' + name
            print(f'  <{newname} grid="g1">\n    ', end=' ', file=xml)
            for x in a:
                print(f'{x!r}', end=' ', file=xml)
            print(f'\n  </{newname}>', file=xml)

        for id, l, u, s, q, in zip(self.id_j, l_j, self.phi_jg, self.phit_jg,
                                   self.pt_jg):
            for name, a in [('ae_partial_wave', u),
                            ('pseudo_partial_wave', s),
                            ('projector_function', q)]:
                print(f'  <{name} state="{id}" grid="g1">\n    ',
                      end=' ', file=xml)
                for x in a:
                    print(f'{x!r}', end=' ', file=xml)
                print(f'\n  </{name}>', file=xml)

        if self.vt_g is not None:
            xml.write('  <pseudo_potential grid="g1">\n')
            for x in self.vt_g:
                print(f'{x!r}', end=' ', file=xml)
            print('\n  </pseudo_potential>', file=xml)

        print('  <kinetic_energy_differences>', end=' ', file=xml)
        nj = len(self.e_kin_jj)
        for j1 in range(nj):
            print('\n    ', end=' ', file=xml)
            for j2 in range(nj):
                print(f'{self.e_kin_jj[j1, j2]!r}', end=' ', file=xml)
        print('\n  </kinetic_energy_differences>', file=xml)

        if self.X_p is not None:
            print('  <exact_exchange_X_matrix>\n    ', end=' ', file=xml)
            for x in self.X_p:
                print(f'{x!r}', end=' ', file=xml)
            print('\n  </exact_exchange_X_matrix>', file=xml)

            print(f'  <exact_exchange core-core="{self.ExxC!r}"/>', file=xml)
            for omega, Ecc in self.ExxC_w.items():
                print(f'  <erfc_exchange omega="{omega}" core-core="{Ecc}"/>',
                      file=xml)
                print(f'  <erfc_exchange_X_matrix omega="{omega}" X_p="',
                      end=' ', file=xml)
                for x in self.X_wp[omega]:
                    print(f'{x!r}', end=' ', file=xml)
                print('"/>', file=xml)

        if self.X_pg is not None:
            print('  <yukawa_exchange_X_matrix>\n    ', end=' ', file=xml)
            for x in self.X_pg:
                print(f'{x!r}', end=' ', file=xml)
            print('\n  </yukawa_exchange_X_matrix>', file=xml)
            print(f'  <yukawa_exchange gamma="{self.X_gamma!r}"/>', file=xml)
        print('</paw_dataset>', file=xml)

    def build(self, xcfunc, lmax, basis, filter=None):
        from gpaw.setup import Setup
        setup = Setup(self, xcfunc, lmax, basis, filter)
        return setup


def search_for_file(name: str, world=None) -> Tuple[str, bytes]:
    """Traverse gpaw setup paths to find file.

    Returns the file path and file contents.  If the file is not
    found, raises RuntimeError."""

    if world is None or world.rank == 0:
        source = b''
        filename = None
        for path in setup_paths:
            pattern = os.path.join(path, name)
            filenames = glob(pattern) + glob(f'{pattern}.gz')
            if filenames:
                # The globbing is a hack to grab the 'newest' version if
                # the files are somehow version numbered; then we want the
                # last/newest of the results (used with SG15).  (User must
                # instantiate (UPF)SetupData directly to override.)
                filename = max(filenames)
                import gzip
                if filename.endswith('.gz'):
                    with gzip.open(filename) as fd:
                        source = fd.read()
                else:
                    source = Path(filename).read_bytes()
                break

    if world is not None:
        if world.rank == 0:
            broadcast((filename, source), 0, world)
        else:
            filename, source = broadcast(None, 0, world)

    if filename is None:
        if name.endswith('basis'):
            _type = 'basis set'
        else:
            _type = 'PAW dataset'
        err = f'Could not find required {_type} file "{name}".'
        helpful_message = """
You need to set the GPAW_SETUP_PATH environment variable to point to
the directories where PAW dataset and basis files are stored.  See
https://wiki.fysik.dtu.dk/gpaw/install.html#install-paw-datasets
for details."""
        raise FileNotFoundError(f'{err}\n{helpful_message}\n')

    return filename, source


class PAWXMLParser(xml.sax.handler.ContentHandler):
    def __init__(self, setup):
        xml.sax.handler.ContentHandler.__init__(self)
        self.setup = setup
        self.id = None
        self.data = None

    def parse(self, source=None, world=None):
        setup = self.setup
        if source is None:
            setup.filename, source = search_for_file(setup.stdfilename, world)

        setup.fingerprint = hashlib.md5(source).hexdigest()

        # XXXX There must be a better way!
        # We don't want to look at the dtd now.  Remove it:
        source = re.compile(b'<!DOCTYPE .*?>', re.DOTALL).sub(b'', source, 1)
        xml.sax.parseString(source, self)

        if setup.zero_reference:
            setup.e_total = 0.0
            setup.e_kinetic = 0.0
            setup.e_electrostatic = 0.0
            setup.e_xc = 0.0

    def startElement(self, name, attrs):
        setup = self.setup
        if name == 'paw_setup' or name == 'paw_dataset':
            setup.version = attrs['version']
            assert [int(v) for v in setup.version.split('.')] >= [0, 4]
        if name == 'atom':
            Z = float(attrs['Z'])
            setup.Z = Z
            assert setup.Z == Z
            setup.Nc = float(attrs['core'])
            Nv = float(attrs['valence'])
            setup.Nv = int(Nv)
            assert setup.Nv == Nv
        elif name == 'xc_functional':
            if attrs['type'] == 'LDA':
                setup.xcname = 'LDA'
            else:
                setup.xcname = attrs['name']
                if attrs['type'] == 'OFDFT':
                    setup.orbital_free = True
                else:
                    assert attrs['type'] == 'GGA'
        elif name == 'ae_energy':
            setup.e_total = float(attrs['total'])
            setup.e_kinetic = float(attrs['kinetic'])
            setup.e_electrostatic = float(attrs['electrostatic'])
            setup.e_xc = float(attrs['xc'])
        elif name == 'core_energy':
            setup.e_kinetic_core = float(attrs['kinetic'])
        elif name == 'state':
            setup.n_j.append(int(attrs.get('n', -1)))
            setup.l_j.append(int(attrs['l']))
            setup.f_j.append(float(attrs.get('f', 0)))
            setup.eps_j.append(float(attrs['e']))
            setup.rcut_j.append(float(attrs.get('rc', -1)))
            setup.id_j.append(attrs['id'])
            # Compatibility with old setups:
            version = [int(v) for v in setup.version.split('.')]
            if version < [0, 6] and setup.f_j[-1] == 0:
                setup.n_j[-1] = -1
        elif name == 'radial_grid':
            if attrs['eq'] == 'r=a*i/(n-i)':
                beta = float(attrs['a'])
                ng = int(attrs['n'])
                setup.rgd = AERadialGridDescriptor(beta / ng, 1.0 / ng, ng)
            elif attrs['eq'] == 'r=a*i/(1-b*i)':
                a = float(attrs['a'])
                b = float(attrs['b'])
                N = int(attrs['n'])
                setup.rgd = AERadialGridDescriptor(a, b, N)
            elif attrs['eq'] == 'r=a*(exp(d*i)-1)':
                a = float(attrs['a'])
                d = float(attrs['d'])
                istart = int(attrs['istart'])
                iend = int(attrs['iend'])
                assert istart == 0
                setup.rgd = AbinitRadialGridDescriptor(a, d, iend + 1)
            else:
                raise ValueError('Unknown grid:' + attrs['eq'])
        elif name == 'shape_function':
            assert attrs['type'] in {'gauss', 'sinc', 'bessel'}
            setup.shape_function = {'type': attrs['type'],
                                    'rc': float(attrs['rc'])}
        elif name in ['ae_core_density', 'pseudo_core_density',
                      'localized_potential', 'yukawa_exchange_X_matrix',
                      'kinetic_energy_differences', 'exact_exchange_X_matrix',
                      'ae_core_kinetic_energy_density',
                      'pseudo_core_kinetic_energy_density',
                      'pseudo_potential']:
            self.data = []
        elif name.startswith('GLLB_'):
            self.data = []
        elif name in ['ae_partial_wave', 'pseudo_partial_wave']:
            self.data = []
            self.id = attrs['state']
        elif name == 'projector_function':
            self.id = attrs['state']
            self.data = []
        elif name == 'erfc_exchange':
            setup.ExxC_w[float(attrs['omega'])] = float(attrs['core-core'])
        elif name == 'exact_exchange':
            setup.ExxC = float(attrs['core-core'])
        elif name == 'erfc_exchange_X_matrix':
            X_p = np.array([float(x) for x in ''.join(attrs['X_p']).split()])
            setup.X_wp[float(attrs['omega'])] = X_p
        elif name == 'yukawa_exchange':
            setup.X_gamma = float(attrs['gamma'])
        elif name == 'core_hole_state':
            setup.has_corehole = True
            setup.fcorehole = float(attrs['removed'])
            setup.lcorehole = 'spdf'.find(attrs['state'][1])
            setup.core_hole_e = float(attrs['eig'])
            setup.core_hole_e_kin = float(attrs['ekin'])
            self.data = []
        elif name == 'zero_potential':
            if 'type' in attrs:
                setup.r0 = float(attrs['r0'])
                setup.nderiv0 = int(attrs['nderiv'])
                if attrs['type'] == 'polynomial':
                    setup.e0 = None
                    setup.l0 = None
                else:
                    setup.e0 = float(attrs['e0'])
                    setup.l0 = 'spdfg'.find(attrs['type'])
            self.data = []
        elif name == 'generator':
            setup.type = attrs['type']
            setup.generator_version = int(attrs.get('version', '1'))
        else:
            self.data = None

    def characters(self, data):
        if self.data is not None:
            self.data.append(data)

    def endElement(self, name):
        setup = self.setup
        if self.data is None:
            return
        x_g = np.array([float(x) for x in ''.join(self.data).split()])
        if name == 'ae_core_density':
            setup.nc_g = x_g
        elif name == 'pseudo_core_density':
            setup.nct_g = x_g
        elif name == 'kinetic_energy_differences':
            setup.e_kin_jj = x_g
        elif name == 'ae_core_kinetic_energy_density':
            setup.tauc_g = x_g
        elif name == 'pseudo_valence_density':
            setup.nvt_g = x_g
        elif name == 'pseudo_core_kinetic_energy_density':
            setup.tauct_g = x_g
        elif name in ['localized_potential', 'zero_potential']:  # XXX
            setup.vbar_g = x_g
        elif name in ['pseudo_potential']:
            setup.vt_g = x_g
        elif name.startswith('GLLB_'):
            # Add setup tags starting with GLLB_ to extra_xc_data. Remove
            # GLLB_ from front of string:
            setup.extra_xc_data[name[5:]] = x_g
        elif name == 'ae_partial_wave':
            j = len(setup.phi_jg)
            assert self.id == setup.id_j[j]
            setup.phi_jg.append(x_g)
        elif name == 'pseudo_partial_wave':
            j = len(setup.phit_jg)
            assert self.id == setup.id_j[j]
            setup.phit_jg.append(x_g)
        elif name == 'projector_function':
            j = len(setup.pt_jg)
            assert self.id == setup.id_j[j]
            setup.pt_jg.append(x_g)
        elif name == 'exact_exchange_X_matrix':
            setup.X_p = x_g
        elif name == 'yukawa_exchange_X_matrix':
            setup.X_pg = x_g
        elif name == 'core_hole_state':
            setup.phicorehole_g = x_g



import numpy as np

from gpaw.utilities import packed_index


class Overlap:
    """Wave funcion overlap of two GPAW objects"""
    def __init__(self, calc):
        self.calc = calc
        self.nb, self.nk, self.ns = self.number_of_states(calc)
        self.gd = self.calc.wfs.gd
        self.kd = self.calc.wfs.kd

    def number_of_states(self, calc):
        # we will need the wave functions
        calc.converge_wave_functions()
        return (calc.get_number_of_bands(), len(calc.get_ibz_k_points()),
                calc.get_number_of_spins())

    def pseudo(self, other, myspin=0, otherspin=0, normalize=True):
        r"""Overlap with pseudo wave functions only

        Parameter
        ---------
        other: gpaw
            gpaw-object containing pseudo wave functions
        normalize: bool
            normalize pseudo wave functions in the overlap integral

        Returns
        -------
        out: array
            u_kij =  \int dx mypsitilde_ki^*(x) otherpsitilde_kj(x)
        """
        nbo, nko, _ = self.number_of_states(other)
        assert self.nk == nko  # XXX allow for different number of k-points ?

        overlap_knn = []
        for k in range(self.nk):
            # XXX what if parallelized over spin or kpoints ?
            kpt_rank, q = self.kd.get_rank_and_index(k)
            assert self.kd.comm.rank == kpt_rank
            kpt_rank, qo = other.wfs.kd.get_rank_and_index(k)
            assert self.kd.comm.rank == kpt_rank

            overlap_nn = np.zeros((self.nb, nbo), dtype=self.calc.wfs.dtype)
            mkpt = self.calc.wfs.kpt_qs[q][myspin]
            okpt = other.wfs.kpt_qs[qo][otherspin]
            psit_nG = mkpt.psit_nG
            norm_n = self.gd.integrate(psit_nG.conj() * psit_nG)
            psito_nG = okpt.psit_nG
            normo_n = other.wfs.gd.integrate(psito_nG.conj() * psito_nG)
            for i in range(self.nb):
                p_nG = np.repeat(psit_nG[i].conj()[np.newaxis], nbo, axis=0)
                overlap_nn[i] = self.gd.integrate(p_nG * psito_nG)
                if normalize:
                    overlap_nn[i] /= np.sqrt(np.repeat(norm_n[i], nbo) *
                                             normo_n)
            overlap_knn.append(overlap_nn)
        return np.array(overlap_knn)

    def full(self, other, myspin=0, otherspin=0):
        """Overlap of Kohn-Sham states including local terms.

        Parameter
        ---------
        other: gpaw
            gpaw-object containing wave functions

        Returns
        -------
        out: array
            u_kij = int dx mypsi_ki^*(x) otherpsi_kj(x)
        """
        ov_knn = self.pseudo(other, normalize=False)
        for k in range(self.nk):
            # XXX what if parallelized over spin or kpoints ?
            kpt_rank, q = self.kd.get_rank_and_index(k)
            assert self.kd.comm.rank == kpt_rank
            kpt_rank, qo = other.wfs.kd.get_rank_and_index(k)
            assert self.kd.comm.rank == kpt_rank

            mkpt = self.calc.wfs.kpt_qs[q][myspin]
            okpt = other.wfs.kpt_qs[qo][otherspin]

            aov_nn = np.zeros_like(ov_knn[k])
            for a, mP_ni in mkpt.P_ani.items():
                oP_ni = okpt.P_ani[a]
                Delta_p = (np.sqrt(4 * np.pi) *
                           self.calc.wfs.setups[a].Delta_pL[:, 0])
                for n0, mP_i in enumerate(mP_ni):
                    for n1, oP_i in enumerate(oP_ni):
                        ni = len(mP_i)
                        assert len(oP_i) == ni
                        for i, mP in enumerate(mP_i):
                            for j, oP in enumerate(oP_i):
                                ij = packed_index(i, j, ni)
                                aov_nn[n0, n1] += Delta_p[ij] * mP.conj() * oP
            self.calc.wfs.gd.comm.sum(aov_nn)
            ov_knn[k] += aov_nn
        return ov_knn



import pytest
from gpaw.mpi import world
import sys

from gpaw.utilities import devnull

from gpaw import GPAW, FermiDirac
from gpaw import KohnShamConvergenceError
from gpaw.utilities import compiled_with_sl
from gpaw.forces import calculate_forces

from ase.build import molecule

# Calculates energy and forces for various parallelizations

pytestmark = pytest.mark.skipif(world.size < 4,
                                reason='world.size < 4')


def test_parallel_fd_parallel_kpt():
    tolerance = 4e-5

    parallel = dict()

    basekwargs = dict(mode='fd',
                      eigensolver='rmm-diis',
                      maxiter=3,
                      # basis='dzp',
                      # nbands=18,
                      nbands=6,
                      kpts=(4, 4, 4),  # 8 kpts in the IBZ
                      parallel=parallel)

    Eref = None
    Fref_av = None

    def run(formula='H2O', vacuum=1.5, cell=None, pbc=1, **morekwargs):
        print(formula, parallel)
        system = molecule(formula)
        kwargs = dict(basekwargs)
        kwargs.update(morekwargs)
        calc = GPAW(**kwargs)
        system.calc = calc
        system.center(vacuum)
        if cell is None:
            system.center(vacuum)
        else:
            system.set_cell(cell)
        system.set_pbc(pbc)

        try:
            system.get_potential_energy()
        except KohnShamConvergenceError:
            pass

        E = calc.hamiltonian.e_total_free
        F_av = calculate_forces(calc.wfs, calc.density,
                                calc.hamiltonian)

        nonlocal Eref, Fref_av
        if Eref is None:
            Eref = E
            Fref_av = F_av

        eerr = abs(E - Eref)
        ferr = abs(F_av - Fref_av).max()

        if calc.wfs.world.rank == 0:
            print('Energy', E)
            print()
            print('Forces')
            print(F_av)
            print()
            print('Errs', eerr, ferr)

        if eerr > tolerance or ferr > tolerance:
            if calc.wfs.world.rank == 0:
                stderr = sys.stderr
            else:
                stderr = devnull
            if eerr > tolerance:
                print('Failed!', file=stderr)
                print('E = %f, Eref = %f' % (E, Eref), file=stderr)
                msg = 'Energy err larger than tolerance: %f' % eerr
            if ferr > tolerance:
                print('Failed!', file=stderr)
                print('Forces:', file=stderr)
                print(F_av, file=stderr)
                print(file=stderr)
                print('Ref forces:', file=stderr)
                print(Fref_av, file=stderr)
                print(file=stderr)
                msg = 'Force err larger than tolerance: %f' % ferr
            print(file=stderr)
            print('Args:', file=stderr)
            print(formula, vacuum, cell, pbc, morekwargs, file=stderr)
            print(parallel, file=stderr)
            raise AssertionError(msg)

    # reference:
    # kpt-parallelization = 8,
    # state-parallelization = 1,
    # domain-decomposition = (1,1,1)
    run()

    # kpt-parallelization = 2,
    # state-parallelization = 2,
    # domain-decomposition = (1,2,1)
    parallel['band'] = 2
    parallel['domain'] = (1, 2, 1)
    run()

    if compiled_with_sl():
        # kpt-parallelization = 2,
        # state-parallelization = 2,
        # domain-decomposition = (1,2,1)
        # with blacs
        parallel['sl_default'] = (2, 2, 2)
        run()

    # perform spin polarization test
    parallel = dict()

    basekwargs = dict(mode='fd',
                      eigensolver='rmm-diis',
                      maxiter=3,
                      nbands=6,
                      kpts=(4, 4, 4),  # 8 kpts in the IBZ
                      parallel=parallel)

    Eref = None
    Fref_av = None

    OH_kwargs = dict(formula='NH2', vacuum=1.5, pbc=1, spinpol=1,
                     occupations=FermiDirac(width=0.1))

    # reference:
    # kpt-parallelization = 4,
    # spin-polarization = 2,
    # state-parallelization = 1,
    # domain-decomposition = (1,1,1)
    run(**OH_kwargs)

    # kpt-parallelization = 2,
    # spin-polarization = 2,
    # domain-decomposition = (1, 2, 1)
    parallel['domain'] = (1, 2, 1)
    run(**OH_kwargs)

    # kpt-parallelization = 2,
    # spin-polarization = 2,
    # state-parallelization = 2,
    # domain-decomposition = (1, 1, 1)
    del parallel['domain']
    parallel['band'] = 2
    run(**OH_kwargs)

    # do last test plus buffer_size keyword
    parallel['buffer_size'] = 150
    run(**OH_kwargs)

    if compiled_with_sl():
        # kpt-parallelization = 2,
        # spin-polarization = 2,
        # state-parallelization = 2,
        # domain-decomposition = (1, 2, 1)
        # with blacs
        del parallel['buffer_size']
        parallel['domain'] = (1, 2, 1)
        parallel['sl_default'] = (2, 1, 2)
        run(**OH_kwargs)

        # kpt-parallelization = 2,
        # state-parallelization = 2,
        # domain-decomposition = (1, 2, 1)
        # with blacs
        parallel['sl_default'] = (2, 2, 2)
        run(**OH_kwargs)

        # do last test plus buffer_size keyword
        parallel['buffer_size'] = 150
        run(**OH_kwargs)



import numpy as np

from gpaw.gauss import Gauss


class Lorentz:
    """Normalized Lorentzian distribution"""
    def __init__(self, width=0.08):
        self.dtype = float
        self.set_width(width)

    def get(self, x, x0=0):
        return self.norm / ((x - x0)**2 + self.width2)

    def set_width(self, width=0.08):
        self.norm = width / np.pi
        self.width2 = width**2
        self._fwhm = 2. * width

    @property
    def fwhm(self):
        return self._fwhm

    @fwhm.setter
    def fwhm(self, value):
        self.set_width(value / 2.)


class Voigt:
    """Voigt profile.

    See http://reference.wolfram.com/language/ref/VoigtDistribution.html
    """
    def __init__(self, width=0.08):
        self.dtype = float
        self.set_width(width)

    def get(self, x, x0=0):
        from scipy.special import erfc
        argm = (-1j * (x - x0) + self.delta) * self.argpre
        argp = (1j * (x - x0) + self.delta) * self.argpre
        res = np.exp(argm**2) * erfc(argm)
        res += np.exp(argp**2) * erfc(argp)
        return res.real * self.prefactor

    def set_width(self, width=0.08):
        """Width is interpreted as [delta, sigma]"""
        if not hasattr(width, '__iter__'):
            width = [width, width]
        self.delta, self.sigma = width
        self.width = np.linalg.norm(np.array(width))
        self.prefactor = 1. / 2 / np.sqrt(2 * np.pi) / self.sigma
        self.argpre = 1. / np.sqrt(2) / self.sigma

        """Full width at half maximum approximation after
        Olivero, J. J.; R. L. Longbothum
        "Empirical fits to the Voigt line width: A brief review"
        J. Quantitative Spectroscopy and Radiative Transfer. 17 (1977) 233-236
        """
        fl = 2 * self.delta
        fg = np.sqrt(8 * np.log(2)) * self.sigma
        self._fwhm = 0.5346 * fl + np.sqrt(0.2166 * fl**2 + fg**2)

    @property
    def fwhm(self):
        return self._fwhm

    @fwhm.setter
    def fwhm(self, value):
        "setting fwhm is not uniquely defined"
        raise NotImplementedError


class ComplexLorentz:
    def __init__(self, width=0.08):
        self.dtype = complex
        self.set_width(width)

    def get(self, x, x0):
        return 0.5 / x0 * (((x + x0) / ((x + x0)**2 + self.width2) -
                            (x - x0) / ((x - x0)**2 + self.width2)) -
                           1.0j * self.width *
                           (1 / ((x + x0)**2 + self.width2) -
                            1 / ((x - x0)**2 + self.width2))
                           )

    def set_width(self, width=0.08):
        self.width = width
        self.width2 = width**2


class ComplexGauss:
    def __init__(self, width=0.08):
        self.dtype = complex
        self.set_width(width)

    def get(self, x, x0):
        from scipy.special import dawsn
        return 0.5 / x0 * (2 * self.wm1 *
                           (dawsn((x + x0) * self.wm1) -
                            dawsn((x - x0) * self.wm1)) -
                           1.0j * self.norm *
                           (np.exp(-((x + x0) * self.wm1)**2) -
                            np.exp(-((x - x0) * self.wm1)**2))
                           )

    def set_width(self, width=0.08):
        self.norm = 1. / width * np.sqrt(np.pi / 2)
        self.wm1 = np.sqrt(.5) / width


class LorentzPole:
    """Pole of a damped harmonic oscillator.

    See: e.g. Frederick Wooten, Optical properties of solids,
    Academic Press Inc. (1972)"""
    def __init__(self, width=0.08, imag=True):
        self.dtype = float
        if imag:
            self.get = self.get_imaginary
        else:
            self.get = self.get_real
        self.set_width(width)

    def get_real(self, x, x0):
        return (x0**2 - x**2) / (
            (x0**2 - x**2)**2 + self.width**2 * x**2)

    def get_imaginary(self, x, x0):
        return self.width * x / (
            (x0**2 - x**2)**2 + self.width**2 * x**2)

    def set_width(self, width=0.08):
        self.width = width


class Folder:
    """Fold a function with normalised Gaussians or Lorentzians

    Example: fold a function y(x) by Lorentzians of 0.2 width

    >>> xlist = np.linspace(-1.0, 1.0, 101)  # list of x-values
    >>> ylist = np.zeros_like(xlist)  # corresponding list of y(x)-values
    >>> ylist[50] = 1.0
    >>> from gpaw.utilities.folder import Folder
    >>> fxlist, fylist = Folder(width=0.2, folding='Lorentz').fold(
    ...     xlist, ylist, dx=0.1, xmin=-1, xmax=1)
    """
    def __init__(self, width,
                 folding='Gauss'):
        self.width = width
        if folding == 'Gauss':
            self.func = Gauss(width)
        elif folding == 'Lorentz':
            self.func = Lorentz(width)
        elif folding == 'ComplexGauss':
            self.func = ComplexGauss(width)
        elif folding == 'ComplexLorentz':
            self.func = ComplexLorentz(width)
        elif folding == 'RealLorentzPole':
            self.func = LorentzPole(width, imag=False)
        elif folding == 'ImaginaryLorentzPole':
            self.func = LorentzPole(width, imag=True)
        elif folding == 'Voigt':
            self.func = Voigt(width)
        else:
            raise RuntimeError('unknown folding "' + folding + '"')

    def fold(self, x, y, dx=None, xmin=None, xmax=None):
        X = np.array(x)
        assert len(X.shape) == 1
        Y = np.array(y)
        assert X.shape[0] == Y.shape[0]

        if xmin is None:
            xmin = np.min(X) - 4 * self.width
        if xmax is None:
            xmax = np.max(X) + 4 * self.width
        if dx is None:
            try:
                dx = self.func.width / 4.
            except AttributeError:
                dx = self.width / 4.

        xl = np.arange(xmin, xmax + 0.5 * dx, dx)

        return self.fold_values(x, y, xl)

    def fold_values(self, x, y, xl=None):
        X = np.array(x)
        assert len(X.shape) == 1
        Y = np.array(y)
        assert X.shape[0] == Y.shape[0]

        if xl is None:
            Xl = np.unique(X)
        else:
            Xl = np.array(xl)
            assert len(Xl.shape) == 1

        # weight matrix
        weightm = np.empty((Xl.shape[0], X.shape[0]),
                           dtype=self.func.dtype)
        for i, x in enumerate(X):
            weightm[:, i] = self.func.get(Xl, x)

        yl = np.tensordot(weightm, Y, axes=(1, 0))

        return Xl, yl



# This test verifies that the FastPoissonSolver produces solutions to
# the Poisson equation with very small residuals for different cells,
# pbcs, and grids.

import itertools
import numpy as np
from ase.build import bulk
from gpaw.poisson import FastPoissonSolver, BadAxesError
from gpaw.grid_descriptor import GridDescriptor
from gpaw.fd_operators import Laplace
from gpaw.mpi import world
from gpaw.utilities import h2gpts

# Test: different pbcs
# For pbc=000, test charged system
# Different cells (orthorhombic/general)
# use_cholesky keyword


def test_poisson_fastpoisson():
    cell_cv = np.array(bulk('Au').cell)
    rng = np.random.RandomState(42)

    tf = range(2)

    def icells():
        # cells: orthorhombic fcc bcc hcp
        yield 'diag', np.diag([3., 4., 5.])

        from ase.build import fcc111
        atoms = fcc111('Au', size=(1, 1, 1))
        atoms.center(vacuum=1, axis=2)
        yield 'fcc111@z', atoms.cell.copy()
        yield 'fcc111@x', atoms.cell[[2, 0, 1]]
        yield 'fcc111@y', atoms.cell[[1, 2, 0]]

        for sym in ['Au', 'Fe', 'Sc']:
            cell = bulk(sym).cell
            yield sym, cell.copy()

    tolerance = 1e-12

    def test(cellno, cellname, cell_cv, idiv, pbc, nn):
        N_c = h2gpts(0.12, cell_cv, idiv=idiv)
        if idiv == 1:
            N_c += 1 - N_c % 2  # We want especially to test uneven grids
        gd = GridDescriptor(N_c, cell_cv, pbc_c=pbc)
        rho_g = gd.zeros()
        phi_g = gd.zeros()
        rho_g[:] = -0.3 + rng.rand(*rho_g.shape)

        # Neutralize charge:
        charge = gd.integrate(rho_g)
        magic = gd.get_size_of_global_array().prod()
        rho_g -= charge / gd.dv / magic
        charge = gd.integrate(rho_g)
        assert abs(charge) < 1e-12

        # Check use_cholesky=True/False ?
        from gpaw.poisson import FDPoissonSolver
        ps = FastPoissonSolver(nn=nn)
        # print('setgrid')

        # Will raise BadAxesError for some pbc/cell combinations
        ps.set_grid_descriptor(gd)

        ps.solve(phi_g, rho_g)

        laplace = Laplace(gd, scale=-1.0 / (4.0 * np.pi), n=nn)

        def get_residual_err(phi_g):
            rhotest_g = gd.zeros()
            laplace.apply(phi_g, rhotest_g)
            residual = np.abs(rhotest_g - rho_g)
            # Residual is not accurate at end of non-periodic directions
            # except for nn=1 (since effectively we use the right stencil
            # only for nn=1 at the boundary).
            #
            # To do this check correctly, the Laplacian should have lower
            # nn at the boundaries.  Therefore we do not test the residual
            # at these ends, only in between, by zeroing the bad ones:
            if nn > 1:
                exclude_points = nn - 1
                for c in range(3):
                    if nn > 1 and not pbc[c]:
                        # get view ehere axis c refers becomes
                        # zeroth dimension:
                        X = residual.transpose(c, (c + 1) % 3, (c + 2) % 3)

                        if gd.beg_c[c] == 1:
                            X[:exclude_points] = 0.0
                        if gd.end_c[c] == gd.N_c[c]:
                            X[-exclude_points:] = 0.0
            return residual.max()

        maxerr = get_residual_err(phi_g)
        pbcstring = '{}{}{}'.format(*pbc)

        if 0:
            ps2 = FDPoissonSolver(relax='J', nn=nn, eps=1e-18)
            ps2.set_grid_descriptor(gd)
            phi2_g = gd.zeros()
            ps2.solve(phi2_g, rho_g)

            phimaxerr = np.abs(phi2_g - phi_g).max()
            maxerr2 = get_residual_err(phi2_g)
            msg = ('{:2d} {:8s} pbc={} err={:8.5e} err[J]={:8.5e} '
                   'err[phi]={:8.5e} nn={:1d}'
                   .format(cellno, cellname, pbcstring, maxerr, maxerr2,
                           phimaxerr, nn))

        state = 'ok' if maxerr < tolerance else 'FAIL'

        msg = ('{:2d} {:8s} grid={} pbc={} err[fast]={:8.5e} nn={:1d} {}'
               .format(cellno, cellname, N_c, pbcstring, maxerr, nn, state))
        if world.rank == 0:
            print(msg)

        return maxerr

    errs = []
    for idiv in [4, 1]:
        for cellno, (cellname, cell_cv) in enumerate(icells()):
            for pbc in itertools.product(tf, tf, tf):
                for nn in [1, 3]:
                    try:
                        err = test(cellno, cellname, cell_cv, idiv, pbc, nn)
                    except BadAxesError:
                        # Ignore incompatible pbc/cell combinations
                        continue

                    errs.append(err)

    for i, err in enumerate(errs):
        assert err < tolerance, err



from __future__ import annotations

import importlib
import os
from types import ModuleType, SimpleNamespace
from typing import Any, Union

import _gpaw
import numpy as np
from ase import Atoms
from ase.calculators.calculator import kpts2sizeandoffsets
from ase.units import Bohr
from gpaw.core import UniformGrid
from gpaw.core.atom_arrays import (AtomArrays, AtomArraysLayout,
                                   AtomDistribution)
from gpaw.core.domain import Domain
from gpaw.gpu.mpi import CuPyMPI
from gpaw.mixer import MixerWrapper, get_mixer_from_keywords
from gpaw.mpi import (MPIComm, Parallelization, serial_comm, synchronize_atoms,
                      world)
from gpaw.new import cached_property, prod
from gpaw.new.basis import create_basis
from gpaw.new.brillouin import BZPoints, MonkhorstPackKPoints
from gpaw.new.density import Density
from gpaw.new.ibzwfs import IBZWaveFunctions
from gpaw.new.input_parameters import InputParameters
from gpaw.new.scf import SCFLoop
from gpaw.new.smearing import OccupationNumberCalculator
from gpaw.new.symmetry import create_symmetries_object
from gpaw.new.xc import XCFunctional
from gpaw.setup import Setups
from gpaw.typing import Array2D, ArrayLike1D, ArrayLike2D
from gpaw.utilities.gpts import get_number_of_grid_points


def builder(atoms: Atoms,
            params: dict[str, Any] | InputParameters) -> DFTComponentsBuilder:
    """Create DFT-components builder.

    * pw
    * lcao
    * fd
    * tb
    * atom
    """
    if isinstance(params, dict):
        params = InputParameters(params)

    mode = params.mode.copy()
    name = mode.pop('name')
    assert name in {'pw', 'lcao', 'fd', 'tb', 'atom'}
    mod = importlib.import_module(f'gpaw.new.{name}.builder')
    name = name.title() if name == 'atom' else name.upper()
    return getattr(mod, f'{name}DFTComponentsBuilder')(atoms, params, **mode)


class DFTComponentsBuilder:
    def __init__(self,
                 atoms: Atoms,
                 params: InputParameters):

        self.atoms = atoms.copy()
        self.mode = params.mode['name']
        self.params = params

        parallel = params.parallel
        world = parallel['world']

        synchronize_atoms(atoms, world)
        self.check_cell(atoms.cell)

        self.initial_magmom_av, self.ncomponents = normalize_initial_magmoms(
            atoms, params.magmoms, params.spinpol or params.hund)

        self.soc = params.soc
        self.nspins = self.ncomponents % 3
        self.spin_degeneracy = self.ncomponents % 2 + 1

        self.xc = self.create_xc_functional()

        self.setups = Setups(atoms.numbers,
                             params.setups,
                             params.basis,
                             self.xc.setup_name,
                             world=world)

        if params.hund:
            c = params.charge / len(atoms)
            for a, setup in enumerate(self.setups):
                self.initial_magmom_av[a, 2] = setup.get_hunds_rule_moment(c)

        symmetries = create_symmetries_object(atoms,
                                              self.setups.id_a,
                                              self.initial_magmom_av,
                                              params.symmetry)
        assert not (self.ncomponents == 4 and len(symmetries) > 1)
        bz = create_kpts(params.kpts, atoms)
        self.ibz = symmetries.reduce(bz, strict=False)

        d = parallel.get('domain', None)
        k = parallel.get('kpt', None)
        b = parallel.get('band', None)
        self.communicators = create_communicators(world, len(self.ibz),
                                                  d, k, b, self.xp)

        if self.mode == 'fd':
            pass  # filter = create_fourier_filter(grid)
            # setups = stups.filter(filter)

        self.nelectrons = self.setups.nvalence - params.charge

        self.nbands = calculate_number_of_bands(params.nbands,
                                                self.setups,
                                                params.charge,
                                                self.initial_magmom_av,
                                                self.mode == 'lcao')
        if self.ncomponents == 4:
            self.nbands *= 2

        self.dtype = params.dtype
        if self.dtype is None:
            if self.ibz.bz.gamma_only:
                self.dtype = float
            else:
                self.dtype = complex
        elif not self.ibz.bz.gamma_only and self.dtype != complex:
            raise ValueError('Can not use dtype=float for non gamma-point '
                             'calculation')

        self.grid, self.fine_grid = self.create_uniform_grids()

        self.fracpos_ac = self.atoms.get_scaled_positions()
        self.fracpos_ac %= 1
        self.fracpos_ac %= 1

    def create_uniform_grids(self):
        raise NotImplementedError

    def create_xc_functional(self):
        return XCFunctional(self.params.xc, self.ncomponents)

    def check_cell(self, cell):
        number_of_lattice_vectors = cell.rank
        if number_of_lattice_vectors < 3:
            raise ValueError(
                'GPAW requires 3 lattice vectors.  '
                f'Your system has {number_of_lattice_vectors}.')

    @cached_property
    def atomdist(self) -> AtomDistribution:
        return AtomDistribution(
            self.grid.ranks_from_fractional_positions(self.fracpos_ac),
            self.grid.comm)

    @cached_property
    def wf_desc(self) -> Domain:
        return self.create_wf_description()

    @cached_property
    def xp(self) -> ModuleType:
        """Array module: Numpy or Cupy."""
        if self.params.parallel['gpu']:
            from gpaw.gpu import cupy, cupy_is_fake
            assert not cupy_is_fake or os.environ.get('GPAW_CPUPY')
            return cupy
        else:
            return np

    def create_wf_description(self) -> Domain:
        raise NotImplementedError

    def __repr__(self):
        return f'{self.__class__.__name__}({self.atoms}, {self.params})'

    @cached_property
    def nct_R(self):
        out = self.grid.empty(xp=self.xp)
        nct_aX = self.get_pseudo_core_densities()
        nct_aX.to_uniform_grid(out=out,
                               scale=1.0 / (self.ncomponents % 3))
        return out

    def create_basis_set(self):
        return create_basis(self.ibz,
                            self.ncomponents % 3,
                            self.atoms.pbc,
                            self.grid,
                            self.setups,
                            self.dtype,
                            self.fracpos_ac,
                            self.communicators['w'],
                            self.communicators['k'])

    def density_from_superposition(self, basis_set):
        return Density.from_superposition(self.grid,
                                          self.nct_R,
                                          self.atomdist,
                                          self.setups,
                                          basis_set,
                                          self.initial_magmom_av,
                                          self.ncomponents,
                                          self.params.charge,
                                          self.params.hund)

    def create_occupation_number_calculator(self):
        return OccupationNumberCalculator(
            self.params.occupations,
            self.atoms.pbc,
            self.ibz,
            self.nbands,
            self.communicators,
            self.initial_magmom_av.sum(0),
            self.ncomponents,
            np.linalg.inv(self.atoms.cell.complete()).T)

    def create_scf_loop(self):
        hamiltonian = self.create_hamiltonian_operator()
        eigensolver = self.create_eigensolver(hamiltonian)

        mixer = MixerWrapper(
            get_mixer_from_keywords(self.atoms.pbc.any(),
                                    self.ncomponents, **self.params.mixer),
            self.ncomponents,
            self.grid._gd,
            world=self.communicators['w'])

        occ_calc = self.create_occupation_number_calculator()
        return SCFLoop(hamiltonian, occ_calc,
                       eigensolver, mixer, self.communicators['w'],
                       {key: value
                        for key, value in self.params.convergence.items()
                        if key != 'bands'},
                       self.params.maxiter)

    def read_ibz_wave_functions(self, reader):
        raise NotImplementedError

    def create_potential_calculator(self):
        raise NotImplementedError

    def read_wavefunction_values(self,
                                 reader,
                                 ibzwfs: IBZWaveFunctions) -> None:
        """ Read eigenvalues, occuptions and projections and fermi levels

        The values are read using reader and set as the appropriate properties
        of (the already instantiated) wavefunctions contained in ibzwfs
        """
        ha = reader.ha

        eig_skn = reader.wave_functions.eigenvalues
        occ_skn = reader.wave_functions.occupations
        P_sknI = reader.wave_functions.projections
        P_sknI = P_sknI.astype(ibzwfs.dtype)

        for wfs in ibzwfs:
            wfs._eig_n = eig_skn[wfs.spin, wfs.k] / ha
            wfs._occ_n = occ_skn[wfs.spin, wfs.k]
            layout = AtomArraysLayout([(setup.ni,) for setup in self.setups],
                                      dtype=self.dtype)
            if self.ncomponents < 4:
                wfs._P_ani = AtomArrays(layout,
                                        dims=(self.nbands,),
                                        data=P_sknI[wfs.spin, wfs.k])
            else:
                wfs._P_ani = AtomArrays(layout,
                                        dims=(self.nbands, 2),
                                        data=P_sknI[wfs.k])

        try:
            ibzwfs.fermi_levels = reader.wave_functions.fermi_levels / ha
        except AttributeError:
            # old gpw-file
            ibzwfs.fermi_levels = np.array(
                [reader.occupations.fermilevel / ha])


def create_communicators(comm: MPIComm = None,
                         nibzkpts: int = 1,
                         domain: Union[int, tuple[int, int, int]] = None,
                         kpt: int = None,
                         band: int = None,
                         xp: ModuleType = np) -> dict[str, MPIComm]:
    parallelization = Parallelization(comm or world, nibzkpts)
    if domain is not None and not isinstance(domain, int):
        domain = prod(domain)
    parallelization.set(kpt=kpt,
                        domain=domain,
                        band=band)
    comms = parallelization.build_communicators()
    comms['w'] = comm

    # We replace size=1 MPI communications with serial_comm so that
    # serial_comm.sum(<cupy-array>) works: XXX
    comms = {key: comm if comm.size > 1 else serial_comm
             for key, comm in comms.items()}

    if xp is not np and not getattr(_gpaw, 'gpu_aware_mpi', False):
        comms = {key: CuPyMPI(comm) for key, comm in comms.items()}

    return comms


def create_fourier_filter(grid):
    gamma = 1.6

    h = ((grid.icell**2).sum(1)**-0.5 / grid.size).max()

    def filter(rgd, rcut, f_r, l=0):
        gcut = np.pi / h - 2 / rcut / gamma
        ftmp = rgd.filter(f_r, rcut * gamma, gcut, l)
        f_r[:] = ftmp[:len(f_r)]

    return filter


def normalize_initial_magmoms(
        atoms: Atoms,
        magmoms: ArrayLike2D | ArrayLike1D | float | None = None,
        force_spinpol_calculation: bool = False) -> tuple[Array2D, int]:
    """Convert magnetic moments to (natoms, 3)-shaped array.

    Also return number of wave function components (1, 2 or 4).

    >>> h = Atoms('H', magmoms=[1])
    >>> normalize_initial_magmoms(h)
    (array([[0., 0., 1.]]), 2)
    >>> normalize_initial_magmoms(h, [[1, 0, 0]])
    (array([[1., 0., 0.]]), 4)
    """
    magmom_av = np.zeros((len(atoms), 3))
    ncomponents = 2

    if magmoms is None:
        magmom_av[:, 2] = atoms.get_initial_magnetic_moments()
    elif isinstance(magmoms, float):
        magmom_av[:, 2] = magmoms
    else:
        magmoms = np.asarray(magmoms)
        if magmoms.ndim == 1:
            magmom_av[:, 2] = magmoms
        else:
            magmom_av[:] = magmoms
            ncomponents = 4

    if (ncomponents == 2 and
        not force_spinpol_calculation and
        not magmom_av[:, 2].any()):
        ncomponents = 1

    return magmom_av, ncomponents


def create_kpts(kpts: dict[str, Any], atoms: Atoms) -> BZPoints:
    if 'kpts' in kpts:
        assert len(kpts) == 1, kpts
        return BZPoints(kpts['kpts'])
    size, offset = kpts2sizeandoffsets(**kpts, atoms=atoms)
    return MonkhorstPackKPoints(size, offset)


def calculate_number_of_bands(nbands: int | str | None,
                              setups: Setups,
                              charge: float,
                              initial_magmom_av: Array2D,
                              is_lcao: bool) -> int:
    nao = setups.nao
    nvalence = setups.nvalence - charge
    M = np.linalg.norm(initial_magmom_av.sum(0))

    orbital_free = any(setup.orbital_free for setup in setups)
    if orbital_free:
        return 1

    if isinstance(nbands, str):
        if nbands == 'nao':
            N = nao
        elif nbands[-1] == '%':
            cfgbands = (nvalence + M) / 2
            N = int(np.ceil(float(nbands[:-1]) / 100 * cfgbands))
        else:
            raise ValueError('Integer expected: Only use a string '
                             'if giving a percentage of occupied bands')
    elif nbands is None:
        # Number of bound partial waves:
        nbandsmax = sum(setup.get_default_nbands()
                        for setup in setups)
        N = int(np.ceil((1.2 * (nvalence + M) / 2))) + 4
        if N > nbandsmax:
            N = nbandsmax
        if is_lcao and N > nao:
            N = nao
    elif nbands <= 0:
        N = max(1, int(nvalence + M + 0.5) // 2 + (-nbands))
    else:
        N = nbands

    if N > nao and is_lcao:
        raise ValueError('Too many bands for LCAO calculation: '
                         f'{nbands}%d bands and only {nao} atomic orbitals!')

    if nvalence < 0:
        raise ValueError(
            f'Charge {charge} is not possible - not enough valence electrons')

    if nvalence > 2 * N:
        raise ValueError(
            f'Too few bands!  Electrons: {nvalence}, bands: {nbands}')

    return N


def create_uniform_grid(mode: str,
                        gpts,
                        cell,
                        pbc,
                        symmetry,
                        h: float = None,
                        interpolation: str = None,
                        ecut: float = None,
                        comm: MPIComm = serial_comm) -> UniformGrid:
    """Create grid in a backwards compatible way."""
    cell = cell / Bohr
    if h is not None:
        h /= Bohr

    realspace = (mode != 'pw' and interpolation != 'fft')
    if not realspace:
        pbc = (True, True, True)

    if gpts is not None:
        size = gpts
    else:
        modeobj = SimpleNamespace(name=mode, ecut=ecut)
        size = get_number_of_grid_points(cell, h, modeobj, realspace,
                                         symmetry.symmetry)
    return UniformGrid(cell=cell, pbc=pbc, size=size, comm=comm)



# Copyright (C) 2003  CAMP
# Please see the accompanying LICENSE file for further information.

import numpy as np
from scipy.special import erf


def I(R, a, b, alpha, beta):  # noqa
    """Calculate integral and derivatives wrt. positions of Gaussian product.

    ::

                /                              2             2
               |  a0+b0  a1+b1  a2+b2  -alpha r   -beta (r-R)
       value = | x      y      z      e          e             dr,
               |
              /

    Returns the tuple (value, d[value]/dx, d[value]/dy, d[value]/dz).
    """
    result = np.zeros(4)
    R = np.array(R)
    result[0] = I1(R, a, b, alpha, beta)
    a = np.array(a)
    for i in range(3):
        a[i] += 1
        result[1 + i] = 2 * alpha * I1(R, tuple(a), b, alpha, beta)
        a[i] -= 2
        if a[i] >= 0:
            result[1 + i] -= (a[i] + 1) * I1(R, tuple(a), b, alpha, beta)
        a[i] += 1
    return result


def I1(R, ap1, b, alpha, beta, m=0):
    if ap1 == (0, 0, 0):
        if b != (0, 0, 0):
            return I1(-R, b, ap1, beta, alpha, m)
        else:
            f = 2 * np.sqrt(np.pi**5 / (alpha + beta)) / (alpha * beta)
            if np.sometrue(R):
                T = alpha * beta / (alpha + beta) * np.dot(R, R)
                f1 = f * erf(T**0.5) * (np.pi / T)**0.5
                if m == 0:
                    return 0.5 * f1
                f2 = f * np.exp(-T) / T**m
                if m == 1:
                    return 0.25 * f1 / T - 0.5 * f2
                if m == 2:
                    return 0.375 * f1 / T**2 - 0.5 * f2 * (1.5 + T)
                if m == 3:
                    return 0.9375 * f1 / T**3 - 0.25 * f2 * (7.5 +
                                                             T * (5 + 2 * T))
                if m == 4:
                    return 3.28125 * f1 / T**4 - 0.125 * f2 * \
                        (52.5 + T * (35 + 2 * T * (7 + 2 * T)))
                if m == 5:
                    return 14.7656 * f1 / T**5 - 0.03125 * f2 * \
                        (945 + T * (630 + T * (252 + T * (72 + T * 16))))
                if m == 6:
                    return 81.2109 * f1 / T**6 - 0.015625 * f2 * \
                        (10395 + T *
                         (6930 + T *
                          (2772 + T * (792 + T * (176 + T * 32)))))
                else:
                    raise NotImplementedError

            return f / (1 + 2 * m)
    for i in range(3):
        if ap1[i] > 0:
            break
    a = ap1[:i] + (ap1[i] - 1,) + ap1[i + 1:]
    result = beta / (alpha + beta) * R[i] * I1(R, a, b, alpha, beta, m + 1)
    if a[i] > 0:
        am1 = a[:i] + (a[i] - 1,) + a[i + 1:]
        result += a[i] / (2 * alpha) * (I1(R, am1, b, alpha, beta, m) -
                                        beta / (alpha + beta) *
                                        I1(R, am1, b, alpha, beta, m + 1))
    if b[i] > 0:
        bm1 = b[:i] + (b[i] - 1,) + b[i + 1:]
        result += b[i] / (2 * (alpha + beta)) * I1(R,
                                                   a, bm1, alpha, beta, m + 1)
    return result


def test_derivatives(R, a, b, alpha, beta, i):
    R = np.array(R)
    a = np.array(a)
    a[i] += 1
    dIdRi = 2 * alpha * I1(R, tuple(a), b, alpha, beta)
    a[i] -= 2
    if a[i] >= 0:
        dIdRi -= (a[i] + 1) * I1(R, tuple(a), b, alpha, beta)
    a[i] += 1
    dr = 0.001
    R[i] += 0.5 * dr
    dIdRi2 = I1(R, tuple(a), b, alpha, beta)
    R[i] -= dr
    dIdRi2 -= I1(R, tuple(a), b, alpha, beta)
    dIdRi2 /= -dr
    R[i] += 0.5 * dr
    return dIdRi, dIdRi2


class Gauss:
    """Normalised Gauss distribution

    from gpaw.gauss import Gauss

    width=0.4
    gs = Gauss(width)

    for i in range(4):
        print('Gauss(i)=', gs.get(i))
    """
    def __init__(self, width=0.08):
        self.dtype = float
        self.set_width(width)

    def get(self, x, x0=0):
        return self.norm * np.exp(-((x - x0) * self.wm1)**2)

    def set_width(self, width=0.08):
        self.norm = 1. / width / np.sqrt(2 * np.pi)
        self.wm1 = np.sqrt(.5) / width
        self._fwhm = np.sqrt(8 * np.log(2)) * width

    @property
    def fwhm(self):
        return self._fwhm

    @fwhm.setter
    def fwhm(self, value):
        self.set_width(value / np.sqrt(8 * np.log(2)))
